from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cmcrameri import cm
from matplotlib.ticker import MultipleLocator

from proteus.atmos_clim.common import read_atmosphere_data
from proteus.interior.wrapper import read_interior_data
from proteus.utils.constants import R_earth
from proteus.utils.plot import get_colour, latex_float, sample_output

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def plot_structure(hf_all: pd.DataFrame, output_dir: str,
                        times: list, int_data:list, atm_data:list,
                        module:str, plot_format: str="pdf"):

    if np.amax(times) < 2:
        log.debug("Insufficient data to make plot_structure")
        return

    log.info("Plot structure")

    if module not in ('spider','aragog'):
        log.debug("Cannot make structure plot with module '%s'"%module)
        return

    # helpfile indicies
    hf_idxs = []
    hf_times = np.array(hf_all["Time"], dtype=float)
    for time in times:
        hf_idxs.append(np.argmin(np.abs(hf_times - time)))

    # helpfile data
    R_int = np.array(hf_all["R_int"], dtype=float)[hf_idxs]
    R_obs = np.array(hf_all["z_obs"], dtype=float)[hf_idxs] + R_int
    R_xuv = np.array(hf_all["z_xuv"], dtype=float)[hf_idxs] + R_int

    # scale to Earth radius
    for r in (R_int, R_obs, R_xuv):
        r /= R_earth

    scale = 0.9
    fig,ax = plt.subplots(1,1, figsize=(10*scale,5*scale))

    norm = mpl.colors.LogNorm(vmin=max(1,times[0]), vmax=times[-1])
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])

    # limits
    r_max = 0.2
    t_max = 1000
    lw = 1.5
    ms=40
    esc_m = '|'
    obs_m = '.'

    # loop over times
    for i,time in enumerate(times):

        # Get atmosphere data for this time
        prof = atm_data[i]
        atm_r = prof["z"]/R_earth + R_int[i]
        atm_t = prof["t"]

        # Get temperuture-depth interior data for this time
        ds = int_data[i]
        if module == "aragog":
            int_t = ds["temp_b"][::-1]
            int_r = ds["radius_b"][::-1] * 1e3
        elif module == "spider":
            int_t = ds.get_dict_values(['data','temp_b'])
            int_r = ds.get_dict_values(['data','radius_b'])
        int_r /= R_earth

        # Plot decorators
        label = latex_float(time)+" yr"
        color = sm.to_rgba(time)

        # Plot atmosphere
        ax.plot(atm_r, atm_t, color=color, label=label, lw=lw, zorder=i)

        # Observed radius
        i_obs = np.argmin(np.abs(atm_r - R_obs[i]))
        ax.scatter(R_obs[i], atm_t[i_obs], color=color, s=ms, marker=obs_m, zorder=i)

        # Escape radius
        i_xuv = np.argmin(np.abs(atm_r - R_xuv[i]))
        ax.scatter(R_xuv[i], atm_t[i_xuv], color=color, s=ms, marker=esc_m, zorder=i)

        # Plot mantle
        R_mantle = np.amax(int_r)
        ax.plot(int_r, int_t,  linestyle='solid',  color=color, lw=lw, zorder=i)

        # Plot core
        R_core = np.amin(int_r)
        ax.plot([0,R_core],[int_t[-1], int_t[-1]], color=color, lw=lw)

        # update limits
        r_max = max(r_max, np.amax(atm_r))
        t_max = max(t_max, np.amax(int_t))
        t_max = max(t_max, np.amax(atm_t))


    r_max *= 1.05
    t_max *= 1.05

    # Decorate plot
    ax.set(xlabel=r"Radius [R$_{\oplus}$]", ylabel="Temperature [K]")
    ax.legend()
    ax.grid(zorder=-1, alpha=0.1)
    leg = ax.legend(framealpha=1.0, loc='upper left')
    ax.add_artist(leg)

    ax.xaxis.set_major_locator(MultipleLocator(0.20))
    ax.xaxis.set_minor_locator(MultipleLocator(0.10))
    ticks = [t for t in ax.get_xticks() if t >= 0]
    ax.set_xticks(ticks, ["%.2f"%t for t in ticks], rotation=-45, ha='center')
    ax.set_xlim(left=0.0, right=r_max)

    ax.yaxis.set_minor_locator(MultipleLocator(500))
    ax.set_ylim(top=0.0, bottom=t_max)

    # Filled regions
    ax.fill_betweenx([0,1e20], x1=0,         x2=R_core,   color=get_colour("cor_bkg"), zorder=-3)
    ax.fill_betweenx([0,1e20], x1=R_core,    x2=R_mantle, color=get_colour("int_bkg"), zorder=-3)
    ax.fill_betweenx([0,1e20], x1=R_mantle,  x2=r_max,    color=get_colour("atm_bkg"), zorder=-3)

    # Line style information
    hdls = []
    hdls.append(ax.scatter(-100, 0, c='k', s=ms, marker=esc_m, label="Escape level"))
    hdls.append(ax.scatter(-100, 0, c='k', s=ms, marker=obs_m, label="Observed level"))
    ax.legend(handles=hdls, loc='lower left', framealpha=1.0, )

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_structure.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_structure_entry(handler: Proteus):
    plot_times,_ = sample_output(handler, extension="_atm.nc", tmin=1e3)
    print("Snapshots:", plot_times)

    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    int_module = handler.config.interior.module
    output_dir = handler.directories['output']

    int_data = read_interior_data(output_dir, int_module, plot_times)
    atm_data = read_atmosphere_data(output_dir, plot_times)

    plot_structure(hf_all,output_dir, plot_times, int_data, atm_data, int_module,
                    plot_format=handler.config.params.out.plot_fmt)


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_structure_entry(handler)



# Generic interior wrapper
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from proteus.outgas.calliope import calc_surface_pressures, calc_target_masses
from proteus.utils.constants import element_list, gas_list

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def calc_target_elemental_inventories(dirs:dict, config:Config, hf_row:dict):
    """
    Calculate total amount of volatile elements in the planet
    """

    if config.outgas.module == 'calliope':
        calc_target_masses(dirs, config, hf_row)
    else:
        raise Exception("Unsupported outgassing module selected!")


def run_outgassing(dirs:dict, config:Config, hf_row:dict):
    '''
    Run outgassing model to get new volatile surface pressures
    '''

    # Floating point errors can be problematic here.
    #    Ensure that zero mass values stay at zero by setting all element mass inventories
    #    which are less than 10 kg equal to zero.
    for e in element_list:
        if hf_row[e + "_kg_total"] < 10.0:
            hf_row[e + "_kg_total"] = 0.0

    # Run outgassing calculation
    if config.outgas.module == 'calliope':
        calc_surface_pressures(dirs, config, hf_row)
    else:
        raise Exception("Unsupported outgassing module selected!")

    # calculate total atmosphere mass (from sum of volatile masses)
    # this will need to be changed when rock vapours are included
    hf_row["M_atm"] = 0.0
    for s in gas_list:
        hf_row["M_atm"] += hf_row[s + "_kg_atm"]



# =============================================================================
# Configuration file of VULCAN used as template in PROTEUS
# =============================================================================


# ====== Setting up paths and filenames for the input and output files  ======
# input:
from __future__ import annotations

use_lowT_limit_rates = False
gibbs_text = 'thermo/gibbs_text.txt' # (all the nasa9 files must be placed in the folder: thermo/NASA9/)
cross_folder = 'thermo/photo_cross/'
com_file = 'thermo/all_compose.txt'
top_BC_flux_file = 'atm/' # the file for the top boundary conditions
bot_BC_flux_file = 'atm/' # the file for the lower boundary conditions
# output:
output_dir = 'output/'
plot_dir = 'plot/'
movie_dir = 'plot/movie/'

# ====== Setting up photochemistry ======
use_photo = True
T_cross_sp = [] # warning: slower start! available atm: 'CO2','H2O','NH3', 'SH','H2S','SO2', 'S2', 'COS', 'CS2'

edd = 0.5 # the Eddington coefficient
dbin1 = 0.1  # the uniform bin width < dbin_12trans (nm)
dbin2 = 2.   # the uniform bin width > dbin_12trans (nm)
dbin_12trans = 240. # the wavelength switching from dbin1 to dbin2 (nm)

# the frequency to update the actinic flux and optical depth
ini_update_photo_frq = 50
final_update_photo_frq = 5

# ====== Setting up ionchemistry ======
use_ion = False
if use_photo is False and use_ion is True:
    print ('Warning: use_ion = True but use_photo = False')
# photoionization needs to run together with photochemistry

# ====== Setting up parameters for the atmosphere ======
rocky = True # for the surface gravity
nz = 100   # number of vertical layers
use_Kzz = True
use_moldiff = True
use_vz = False
Kzz_prof = 'Pfunc' # Options: 'const','file' or 'Pfunc' (Kzz increased with P^-0.4)
K_max = 1e5        # for Kzz_prof = 'Pfunc'
K_p_lev = 0.1      # for Kzz_prof = 'Pfunc'
vz_prof = 'const'  # Options: 'const' or 'file'
const_Kzz = 1.E10 # (cm^2/s) Only reads when use_Kzz = True and Kzz_prof = 'const'
const_vz = 0 # (cm/s) Only reads when use_vz = True and vz_prof = 'const'


# frequency for updating dz and dzi due to change of mu
update_frq = 50

# ====== Setting up the boundary conditions ======
# Boundary Conditions:
use_topflux = False
use_botflux = False
max_flux = 1e13  # upper limit for the diffusion-limit fluxes

# ====== Reactions to be switched off  ======
remove_list = [] # in pairs e.g. [1,2]

# == Condensation ======
use_condense = False
use_settling = False
condense_sp = []
non_gas_sp = []
fix_species = []      # fixed the condensable species after condensation-evapoation EQ has reached

# ====== steady state check ======
st_factor = 0.5
conv_step = 500

# ====== Setting up numerical parameters for the ODE solver ======
ode_solver = 'Ros2' # case sensitive
use_print_prog = True
use_print_delta = False
print_prog_num = 200  # print the progress every x steps
dttry = 1.E-10
trun_min = 1e2
runtime = 1.E22
dt_min = 1.E-14
dt_max = runtime*1e-5
dt_var_max = 2.
dt_var_min = 0.5
count_min = 120
count_max = int(1E5)

mtol_conv = 1.E-16
pos_cut = 0
nega_cut = -1.
loss_eps = 1e12 # for using BC
yconv_cri = 0.01 # for checking steady-state
slope_cri = 1.e-4
yconv_min = 0.1
flux_cri = 0.1
flux_atol = 1. # the tol for actinc flux (# photons cm-2 s-1 nm-1)

# ====== Setting up numerical parameters for Ros2 ODE solver ======
atol = 7.E-3             # Try decreasing this if the solutions are not stable
mtol = 1.E-22
rtol = 0.4              # relative tolerence for adjusting the stepsize

# ====== Setting up for ouwtput and plotting ======
plot_TP = False
use_live_flux = False
use_plot_end = False
use_plot_evo = False
use_save_movie = False
use_flux_movie = False
plot_height = False
use_PIL = True
live_plot_frq = 50
save_movie_rate = live_plot_frq
y_time_freq = 1  #  storing data for every 'y_time_freq' step
# output:
output_humanread = False
use_shark = False
save_evolution = False   # save the evolution of chemistry (y_time and t_time) for every save_evo_frq step
save_evo_frq = 50

use_live_plot  = False
scat_sp = []
diff_esc = []
plot_spec = ['H2','H','H2O','OH','CH4','HCN','N2','NH3']
atm_type = 'file'

# ========== PARAMETERS BELOW ARE INSERTED BY OFFLINE_CHEMISTRY AT RUNTIME ===========



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from proteus.utils.plot import get_colour

if TYPE_CHECKING:
    from proteus import Proteus
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def plot_fluxes_global(hf_all:pd.DataFrame, output_dir: str, config: Config, t0: float=100.0):

    # Get values
    hf_crop = hf_all.loc[hf_all["Time"]>t0]
    time = np.array(hf_crop["Time"])
    if len(time) < 3:
        log.warning("Cannot make plot with less than 3 samples")
        return

    log.info("Plot global fluxes")

    F_net = np.array(hf_crop["F_atm"])
    F_asf = np.array(hf_crop["F_ins"]) * config.orbit.s0_factor * (1.0 - config.atmos_clim.albedo_pl) * np.cos(config.orbit.zenith_angle * np.pi/180.0)
    F_olr = np.array(hf_crop["F_olr"])
    F_upw = np.array(hf_crop["F_olr"]) + np.array(hf_crop["F_sct"])
    F_int = np.array(hf_crop["F_int"])
    F_tidal = np.array(hf_crop["F_tidal"])
    F_radio = np.array(hf_crop["F_radio"])

    # Create plot
    mpl.use('Agg')
    scale = 1.2
    fig,ax = plt.subplots(1,1,figsize=(6*scale,5*scale))
    lw = 2.0
    al = 0.96

    # Steam runaway line
    ax.axhline(y=280.0, color='black', lw=lw, linestyle='dashed', label="S-N limit", zorder=1)

    # Plot fluxes
    ax.plot(time, F_tidal, lw=lw, alpha=al, color=get_colour("tidal"),label="Tidal")
    ax.plot(time, F_radio, lw=lw, alpha=al, color=get_colour("radio"),label="Radio.")
    ax.plot(time, F_int,   lw=lw, alpha=al, color=get_colour("int"),  label="Net (int.)")
    ax.plot(time, F_net,   lw=lw, alpha=al, color=get_colour("atm"),  label="Net (atm.)")
    ax.plot(time, F_olr,   lw=lw, alpha=al, color=get_colour("OLR"),  label="OLR")
    ax.plot(time, F_upw,   lw=lw, alpha=al, color=get_colour("sct"),  label="OLR + Scat.")
    ax.plot(time, F_asf,   lw=lw, alpha=al, color=get_colour("ASF"),  label="ASF", linestyle='dotted')


    # Configure plot
    ax.set_yscale("symlog")
    ax.set_ylabel("Unsigned flux [W m$^{-2}$]")
    ax.set_xscale("log")
    ax.set_xlabel("Time [yr]")
    ax.set_xlim(time[0], time[-1])
    legend = ax.legend(loc='lower left')
    legend.get_frame().set_alpha(None)
    legend.get_frame().set_facecolor((1, 1, 1, 0.99))
    ax.grid(color='black', alpha=0.05)

    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_fluxes_global.%s" % config.params.out.plot_fmt,
                bbox_inches='tight', dpi=200)


def plot_fluxes_global_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plot
    plot_fluxes_global(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        config=handler.config,
    )


if __name__ == '__main__':
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_fluxes_global_entry(handler)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import cmcrameri as cm
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from proteus.utils.plot_offchem import offchem_read_grid, offchem_slice_grid

if TYPE_CHECKING:
    from proteus import Proteus

mpl.use("Agg")


def plot_offchem_grid_cross(
        grid_dir:str,
        x_var:str,
        y_var:str,
        z_var:str,
        cvar_dict:dict={},
        contour:bool=True,
    labelcontrols:bool=False,
    ):
    """Plot a cross-section of the GridOfflineChemistry output.

    Parameters
    ----------
    grid_dir : str
        Path to grid output folder
    x_var : str
        Independent variable on the x-axis
    y_var : str
        Independent variable on the y-axis
    z_var : str
        Dependent variable on the z-axis (colour bar)
    cvar_dict : dict
        Control variable filter dictionary (see `offchem_slice_grid` for more info)
    contour : bool
        Plot using contours instead of scatter points.
    labelcontrols : bool
        Write the control variables on the plot

    """

    # Read grid
    years, opts, data  = offchem_read_grid(grid_dir)

    # Validate x_var and y_var choices
    for v in [x_var, y_var]:
        if v not in opts[0].keys():
            raise Exception("Invalid independent variable '%s'" % v)
        try:
            _ = float(opts[0][v])
        except ValueError:
            raise Exception("Independent variable is not a scalar quantity")

    # Slice grid
    if len(cvar_dict.keys()) > 0:
        years, opts, data = offchem_slice_grid(years, opts, data, cvar_dict)
    gpoints = np.shape(opts)[0]

    if contour and (gpoints < 12):
        print("WARNING: grid size is small but contour plotting is enabled - expect ugly plots")

    # Flatten x,y,z data into 1D arrays...

    x_arr = []
    y_arr = []
    z_arr = []

    x_scale = 'linear'
    y_scale = 'linear'
    z_scale = 'linear'
    cb_vmin = None
    cb_vmax = None

    for gi in range(gpoints):

        # Get data at last time-sample for now (think about this later)

        y = years[gi][-1]
        d = data[gi][-1]

        x_arr.append(float(opts[gi][x_var]))
        y_arr.append(float(opts[gi][y_var]))

        # handle possible z_var values
        z_this = None

        if z_var == 'p_toa':
            z_this = d['pressure'][-1]
            z_scale = 'log'
            z_label = 'Total pressure (TOA) [bar]'

        elif z_var == 'p_boa':
            z_this = d['pressure'][0]
            z_scale = 'log'
            z_label = 'Total pressure (BOA) [bar]'

        elif z_var == 'T_surf':
            z_this = d['temperature'][0]
            z_label = 'Surface temperature [K]'

        elif z_var[0] == 'x': # mole fraction cases

            prf = z_var[:6]
            sp = z_var[6:]
            k  = 'mx_' + sp

            if k in d.keys():
                if prf == 'x_toa_': # mixing ratio from VULCAN (TOA)
                        z_this = d[k][0]
                        z_scale = 'log'
                        z_label = '%s kinetic mole fraction (TOA)'

                elif prf == 'x_boa_': # mixing ratio from VULCAN (BOA)
                        z_this = d[k][-1]
                        z_scale = 'log'
                        z_label = '%s kinetic mole fraction (BOA)' % sp

                elif prf == 'x_avg_': # mixing ratio from VULCAN (column average)
                        z_this = np.average(d[k][:])
                        z_scale = 'log'
                        z_label = '%s kinetic mole fraction (column average)' % sp

                elif prf == 'x_run_': # runtime well-mixed mole fraction
                        z_this = d[k]
                        z_scale = 'log'
                        z_label = '%s MVE mole fraction' % sp

        elif z_var == 'lifetime': # magma ocean lifetime
            z_this = y
            z_label = 'Magma ocean liftime [yr]'

        else:
            raise Exception("Invalid dependent variable '%s'" % z_var)

        if z_this is None:
            raise Exception("Could not read dependent variable '%s' from grid" % z_var)

        z_arr.append(float(z_this))

    # Plot data
    fig,ax = plt.subplots(1,1)

    ax.set_ylabel(y_var)
    ax.set_yscale(y_scale)

    ax.set_xlabel(x_var)
    ax.set_xscale(x_scale)

    if cb_vmin is None:
        cb_vmin = np.amin(z_arr)

    if cb_vmax is None:
        cb_vmax = np.amax(z_arr)

    if z_scale == 'log':
        norm = mpl.colors.LogNorm(vmin=cb_vmin, vmax=cb_vmax)
    else:
        norm = mpl.colors.Normalize(vmin=cb_vmin, vmax=cb_vmax)

    cmap = cm.batlowK_r
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)

    cbar = fig.colorbar(sm, cax=cax, orientation='vertical')
    cbar.ax.set_ylabel(z_label)

    if contour:
         ax.tricontourf(x_arr, y_arr, z_arr, cmap=cmap, norm=norm, levels=100)
    else:
        ax.scatter(x_arr, y_arr, c=z_arr, cmap=cmap, norm=norm)

    if labelcontrols:
        lbl = "Control variables:\n"
        for k in cvar_dict.keys():
             lbl += k + " = %g\n" % cvar_dict[k]
        ax.text(0.02, 0.98, lbl,
                horizontalalignment='left',
                verticalalignment='top',
                transform = ax.transAxes,
                color='white',
                backgroundcolor=(0.0, 0.0, 0.0, 0.3),
                fontsize='small')

    fig.tight_layout()
    fig.savefig(grid_dir+"/plot_offchem_grid_cross_[%s]_[%s]_[%s].pdf"%(x_var, y_var, z_var))
    plt.close('all')


def plot_offchem_grid_cross_entry(handler: Proteus):
    raise NotImplementedError(
        "`plot_offchem_grid_cross()` does not support command-line execution."
    )


if __name__ == '__main__':
    print("Plotting offchem grid (crossection)...")

    # Folder
    fol = "output/pspace_redox/"

    # Independent vars
    xv  = "fO2_shift_IW"
    yv  = "CH_ratio"

    # Dependent var
    zv  = "T_surf"

    # Filter
    cf = {"semimajoraxis":0.01154}

    # Plotting options
    contour = False
    lctrl   = True

    print("    x: %s" % xv)
    print("    y: %s" % yv)
    print("    z: %s" % zv)
    print("    f: %s" % cf)

    plot_offchem_grid_cross(os.path.abspath(fol), xv, yv, zv, cf, contour, lctrl)

    print("Done!")



from __future__ import annotations

import sys
from pathlib import Path

import click

from .proteus import Proteus

config_option = click.option(
    '-c',
    '--config',
    'config_path',
    type=click.Path(exists=True, dir_okay=False, path_type=Path, resolve_path=True),
    help='Path to config file',
    required=True,
)


@click.group()
@click.version_option(package_name='fwl-proteus')
def cli():
    pass


def list_plots(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    from .plot import plot_dispatch

    click.echo(' '.join(plot_dispatch))
    sys.exit()


@click.command()
@click.argument('plots', nargs=-1)
@config_option
@click.option(
    '-l',
    '--list',
    is_flag=True,
    default=False,
    help='List available plots and exit',
    is_eager=True,
    expose_value=False,
    callback=list_plots,
)
def plot(plots: str, config_path: Path):
    """(Re-)generate plots from completed run"""
    from .plot import plot_dispatch

    click.echo(f'Config: {config_path}')

    handler = Proteus(config_path=config_path)

    for plot in plots:
        click.echo(f'Plotting: {plot}')
        plot_func = plot_dispatch[plot]
        plot_func(handler=handler)


@click.command()
@config_option
@click.option(
    '-r',
    '--resume',
    is_flag=True,
    default=False,
    help='Resume simulation from disk',
)
@click.option(
    '-o',
    '--offline',
    is_flag=True,
    default=False,
    help='Run in offline mode; do not connect to the internet',
)
def start(config_path: Path, resume: bool, offline: bool):
    """Start proteus run"""
    runner = Proteus(config_path=config_path)
    runner.start(resume=resume, offline=offline)


cli.add_command(plot)
cli.add_command(start)


@click.group()
def get():
    """Get data and modules"""
    pass


@click.command()
@click.option('-n', '--name',  'name',  type=str, help='Name of spectral file group')
@click.option('-b', '--bands', 'bands', type=str, help='Number of bands')
def spectral(**kwargs):
    """Get spectral files

    By default, download all files.
    """
    from .utils.data import download_spectral_file
    download_spectral_file(kwargs["name"],kwargs["bands"])


@click.command()
def stellar():
    """Get stellar spectra"""
    from .utils.data import download_evolution_tracks, download_stellar_spectra

    for track in ["Spada","Baraffe"]:
        download_evolution_tracks(track)
    download_stellar_spectra()

@click.command()
def surfaces():
    """Get surface albedos"""
    from .utils.data import download_surface_albedos
    download_surface_albedos()

@click.command()
def reference():
    """Get reference data (exoplanet populations, mass-radius curves, etc.)"""
    from .utils.data import download_exoplanet_data, download_massradius_data
    download_exoplanet_data()
    download_massradius_data()


@click.command()
def socrates():
    """Set up SOCRATES"""
    from .utils.data import get_socrates
    get_socrates()

@click.command()
def petsc():
    """Set up PETSc"""
    from .utils.data import get_petsc
    get_petsc()

@click.command()
def spider():
    """Set up SPIDER"""
    from .utils.data import get_spider
    get_spider()


cli.add_command(get)
get.add_command(spectral)
get.add_command(surfaces)
get.add_command(reference)
get.add_command(stellar)
get.add_command(socrates)
get.add_command(petsc)
get.add_command(spider)


@click.command()
def doctor():
    """Diagnose your PROTEUS installation"""
    from .doctor import doctor_entry
    doctor_entry()


cli.add_command(doctor)

if __name__ == '__main__':
    cli()



from __future__ import annotations

from pathlib import Path

import cattrs
import tomllib

from ._config import Config


def read_config(path: Path | str) -> dict:
    """Read config file from path"""
    with open(path, 'rb') as f:
        config = tomllib.load(f)
    return config

def read_config_object(path: Path | str) -> Config:
    """Read and validate config into Config object."""
    cfg = read_config(path)
    return cattrs.structure(cfg, Config)

__all__ = [
    'Config',
    'read_config_object',
    'read_config'
]



# Contains routines for setting the model timestep
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pandas as pd

from proteus.utils.helper import UpdateStatusfile

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def next_step(config:Config, dirs:dict, hf_row:dict, hf_all:pd.DataFrame, step_sf:float) -> float:

    # Time stepping adjustment
    if hf_row["Time"] < 2.0:
        # First year, use small step
        dtswitch = 1.0
        log.info("Time-stepping intent: static")

    else:
        if config.params.dt.method == 'proportional':
            # Proportional time-step calculation
            log.info("Time-stepping intent: proportional")
            dtswitch = hf_row["Time"] / config.params.dt.proportional.propconst

        elif config.params.dt.method == 'adaptive':
            # Dynamic time-step calculation

            # Try to maintain a minimum step size of dt_initial at first
            if hf_row["Time"] > config.params.dt.initial:
                dtprev = float(hf_all.iloc[-1]["Time"] - hf_all.iloc[-2]["Time"])
            else:
                dtprev = config.params.dt.initial
            log.debug("Previous step size: %.2e yr"%dtprev)

            # Change in F_int
            F_int_2  = hf_all.iloc[-2]["F_int"]
            F_int_1  = hf_all.iloc[-1]["F_int"]
            F_int_12 = abs(F_int_1 - F_int_2)

            # Change in F_atm
            F_atm_2  = hf_all.iloc[-2]["F_atm"]
            F_atm_1  = hf_all.iloc[-1]["F_atm"]
            F_atm_12 = abs(F_atm_1 - F_atm_2)

            # Change in global melt fraction
            phi_2  = hf_all.iloc[-2]["Phi_global"]
            phi_1  = hf_all.iloc[-1]["Phi_global"]
            phi_12 = abs(phi_1 - phi_2)

            # Determine new time-step given the tolerances
            dt_rtol = config.params.dt.adaptive.rtol
            dt_atol = config.params.dt.adaptive.atol
            speed_up = True
            speed_up = speed_up and ( F_int_12 < dt_rtol*abs(F_int_2) + dt_atol )
            speed_up = speed_up and ( F_atm_12 < dt_rtol*abs(F_atm_2) + dt_atol )
            speed_up = speed_up and ( phi_12   < dt_rtol*abs(phi_2  ) + dt_atol )

            if speed_up:
                dtswitch = dtprev * 1.1
                log.info("Time-stepping intent: speed up")
            else:
                dtswitch = dtprev * 0.75
                log.info("Time-stepping intent: slow down")


        elif config.params.dt.method == 'maximum':
            # Always use the maximum time-step, which can be adjusted in the cfg file
            log.info("Time-stepping intent: maximum")
            dtswitch = config.params.dt.maximum

        else:
            UpdateStatusfile(dirs, 20)
            raise Exception(f"Invalid time-stepping method: {config.params.dt.method}")

        # Step scale factor (is always <= 1.0)
        dtswitch *= step_sf

        # Step-size ceiling
        dtswitch = min(dtswitch, config.params.dt.maximum )    # Absolute

        # Step-size floor
        dtswitch = max(dtswitch, hf_row["Time"]*0.0001)     # Relative
        dtswitch = max(dtswitch, config.params.dt.minimum )    # Absolute

    log.info("New time-step is %1.2e years" % dtswitch)
    return dtswitch



# Functions used to handle escape
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from zephyrus.escape import EL_escape

from proteus.utils.constants import AU, element_list, ergcm2stoWm2, secs_per_year
from proteus.utils.helper import PrintHalfSeparator

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def RunEscape(config:Config, hf_row:dict, dt:float, stellar_track):
    """Run Escape submodule.

    Generic function to run escape calculation using ZEPHYRUS or dummy.

    Parameters
    ----------
        config : dict
            Dictionary of configuration options
        hf_row : dict
            Dictionary of helpfile variables, at this iteration only
        dt : float
            Time interval over which escape is occuring [yr]
        stellar_track : mors star object
            Mors star object storing spada track data.
    """

    PrintHalfSeparator()

    if not config.escape.module:
        # solvevol_target is undefined?
        pass
    elif config.escape.module == 'zephyrus':
        hf_row["esc_rate_total"] = RunZEPHYRUS(config, hf_row, stellar_track)
    elif config.escape.module == 'dummy':
        hf_row["esc_rate_total"] = config.escape.dummy.rate
    else:
        raise ValueError(f"Invalid escape model: {config.escape.module}")

    log.info(
        "Bulk escape rate: %.2e kg yr-1 = %.2e kg s-1"
        % (hf_row["esc_rate_total"] * secs_per_year, hf_row["esc_rate_total"])
        )

    solvevol_target = SpeciesEscapeFromTotalEscape(hf_row, dt)

    # store in hf_row as elements
    for e in element_list:
        if e == "O":
            continue
        hf_row[e + "_kg_total"] = solvevol_target[e]

def RunZEPHYRUS(config, hf_row, stellar_track):
    """Run energy-limited escape (for now) model.

    Parameters
    ----------
        config : dict
            Dictionary of configuration options
        hf_row : dict
            Dictionary of helpfile variables, at this iteration only
        stellar_track : mors star object
            Mors star object storing spada track data.
    Returns
    -------
        mlr : float
            Bulk escape rate [kg s-1]
    """

    log.info("Running EL escape (ZEPHYRUS) ...")

    # Get the age of the star at time t to compute XUV flux at that time
    age_star = hf_row["age_star"] / 1e6 # [Myrs]

    # Interpolating the XUV flux at the age of the star
    Fxuv_star_SI = ((stellar_track.Value(age_star, 'Lx') + stellar_track.Value(age_star, 'Leuv'))
                             / (4 * np.pi * (config.orbit.semimajoraxis * AU * 1e2)**2)) * ergcm2stoWm2

    log.info(f"Interpolated Fxuv_star_SI at age_star = {age_star} Myr is {Fxuv_star_SI} [W/m2]")

    # Compute energy-limited escape
    mlr = EL_escape(config.escape.zephyrus.tidal, #tidal contribution (True/False)
                    config.orbit.semimajoraxis * AU, #planetary semi-major axis [m]
                    config.orbit.eccentricity, #eccentricity
                    hf_row["M_planet"], #planetary mass [kg]
                    config.star.mass, #stellar mass [kg]
                    config.escape.zephyrus.efficiency, #efficiency factor
                    hf_row["R_int"], #planetary radius [m]
                    hf_row["R_xuv"], #XUV optically thick planetary radius [m]
                    Fxuv_star_SI)   # [kg s-1]

    return mlr

def SpeciesEscapeFromTotalEscape(hf_row:dict, dt:float):

    out = {}
    esc = hf_row["esc_rate_total"]

    # calculate total mass of volatiles (except oxygen, which is set by fO2)
    M_vols = 0.0
    for e in element_list:
        if e=='O':
            continue
        M_vols += hf_row[e+"_kg_total"]

    # for each elem, calculate new total inventory while
    # maintaining a constant mass mixing ratio

    for e in element_list:
        if e=='O':
            continue

        # current elemental mass ratio in total
        emr = hf_row[e+"_kg_total"]/M_vols

        log.debug("    %s mass ratio = %.2e "%(e,emr))

        # new total mass of element e, keeping a constant mixing ratio of that element
        out[e] = emr * (M_vols - esc * dt * secs_per_year)

        # do not allow negative masses
        out[e] = max(0.0, out[e])

    return out



# Small physics functions that can be used universally
# This file should not depend on too many other files, as this can cause circular import issues

from __future__ import annotations

import logging

import numpy as np

from proteus.utils.constants import const_c, const_h, const_k

log = logging.getLogger("fwl."+__name__)

def planck_wav(tmp:float, wav:float):
    '''
    Evaluate the planck function at a given temperature and wavelength.
    All in SI units.

    Parameters
    -----------
        tmp : float
            Temperature [K]
        wav : float
            Wavelength [m]

    Returns
    -----------
        flx : float
            Spectral flux density [W m-2 m-1]
    '''

    # Optimisation variables
    wav5 = wav*wav*wav*wav*wav
    hc   = const_h * const_c

    # Calculate planck function value [W m-2 sr-1 m-1]
    # http://spiff.rit.edu/classes/phys317/lectures/planck.html
    with np.errstate(all='raise'):
        # Catch overflow error, which can occur when evaluating at short wavelengths
        try:
            flx = 2.0 * hc * (const_c / wav5) / ( np.exp(hc / (wav * const_k * tmp)) - 1.0)
        except FloatingPointError:
            flx = 0.0

    # Integrate solid angle (hemisphere)
    flx = flx * np.pi

    # Do not allow zero or near-zero fluxes
    flx = max(flx, 1e-40)

    return flx



from __future__ import annotations

from itertools import chain

import pytest
from helpers import PROTEUS_ROOT

from proteus import Proteus
from proteus.config import Config, read_config_object
from proteus.config._converters import none_if_none
from proteus.config._delivery import Volatiles
from proteus.config._outgas import Calliope

PATHS = chain(
    (PROTEUS_ROOT / 'input').glob('*.toml'),
)


@pytest.fixture
def cfg():
    path = PROTEUS_ROOT / 'input' / 'demos'/ 'dummy.toml'
    obj = read_config_object(path)

    assert isinstance(obj, Config)

    return obj


def test_none_if_none():
    assert none_if_none('none') is None

    for val in ('', 'notnone', 'mors', 'None'):
        assert none_if_none(val) is not None


@pytest.mark.parametrize('path', PATHS)
def test_proteus_init(path):
    runner = Proteus(config_path=path)

    assert isinstance(runner.config, Config)


def test_calliope_is_included():
    conf = Calliope(
        T_floor=0.1,
        include_CO=False,
        include_H2=True,
    )
    assert not conf.is_included('CO')
    assert conf.is_included('H2')
    with pytest.raises(AttributeError):
        conf.is_included('fails')


def test_delivery_get_pressure():
    conf = Volatiles(
        N2 = 123.0
    )
    assert conf.get_pressure('N2') == 123.0
    with pytest.raises(AttributeError):
        conf.get_pressure('fails')



from __future__ import annotations

from attr.validators import ge, gt, in_
from attrs import define, field

from ._converters import none_if_none


@define
class Elements:
    """Initial volatile inventory by planetary element abundances.

    For hydrogen: only H_oceans or H_ppmw should be used at any one time.
    For X in {C, N, S}: only XH_ratio or X_ppmw should be used at any one time.

    Attributes
    ----------
    H_oceans: float
        Absolute hydrogen inventory, units of equivalent Earth oceans.
    H_ppmw: float
        Relative hydrogen inventory, ppmw relative to mantle mass.

    CH_ratio: float
        Carbon metallicity. C/H mass ratio in combined mantle+atmosphere system.
    C_ppmw: float
        Relative carbon inventory, ppmw relative to mantle mass.

    NH_ratio: float
        Nitrogen metallicity. N/H mass ratio in combined mantle+atmosphere system.
    N_ppmw: float
        Relative nitrogen inventory, ppmw relative to mantle mass.

    SH_ratio: float
        Sulfur metallicity. C/H mass ratio in combined mantle+atmosphere system.
    S_ppmw: float
        Absolute sulfur inventory, ppmw relative to mantle mass.
    """
    H_oceans: float = field(default=0.0, validator=ge(0))
    H_ppmw: float   = field(default=0.0, validator=ge(0))

    CH_ratio: float = field(default=0.0, validator=ge(0))
    C_ppmw: float   = field(default=0.0, validator=ge(0))

    NH_ratio: float = field(default=0.0, validator=ge(0))
    N_ppmw: float   = field(default=0.0, validator=ge(0))

    SH_ratio: float = field(default=0.0, validator=ge(0))
    S_ppmw: float   = field(default=0.0, validator=ge(0))


@define
class Volatiles:
    """Initial volatile inventory set by partial pressures in atmosphere.

    Attributes
    ----------
    H2O: float
        Initial atmospheric partial surface pressure of H2O [bar].
    CO2: float
        Initial atmospheric partial surface pressure of CO2 [bar].
    N2: float
        Initial atmospheric partial surface pressure of N2 [bar].
    S2: float
        Initial atmospheric partial surface pressure of S2 [bar].
    SO2: float
        Initial atmospheric partial surface pressure of SO2 [bar].
    H2: float
        Initial atmospheric partial surface pressure of H2 [bar].
    CH4: float
        Initial atmospheric partial surface pressure of CH4 [bar].
    CO: float
        Initial atmospheric partial surface pressure of CO [bar].
    """
    H2O: float = field(default=0, validator=ge(0))
    CO2: float = field(default=0, validator=ge(0))
    N2: float = field(default=0, validator=ge(0))
    S2: float = field(default=0, validator=ge(0))
    SO2: float = field(default=0, validator=ge(0))
    H2: float = field(default=0, validator=ge(0))
    CH4: float = field(default=0, validator=ge(0))
    CO: float = field(default=0, validator=ge(0))

    def get_pressure(self, s: str) -> float:
        """Helper method for getting the pressure for `vol` by string."""
        return getattr(self, s)

@define
class Delivery:
    """Initial volatile inventory, radionuclide concentration, and delivery model selection.

    Attributes
    ----------
    initial: str
        Method by which to set the initial volatile inventory to use. Options: 'volatiles', 'elements'.
    module: str
        Delivery module to use (Not used as of yet).
    elements: Elements
        Parameters used when setting volatile inventory by element abundances.
    volatiles: Volatiles
        Parameters used when setting volatile inventory by partial pressures.
    radio_tref: float
        Reference age for setting radioactive decay [Gyr].
    radio_U: float
        Concentration (ppmw) of uranium at reference age of `t=radio_tref`
    radio_K: float
        Concentration (ppmw) of potassium at reference age of `t=radio_tref`
    radio_Th: float
        Concentration (ppmw) of thorium at reference age of `t=radio_tref`
    """
    initial: str = field(validator=in_(('elements', 'volatiles')))

    module: str | None = field(validator=in_((None,)), converter=none_if_none)

    elements: Elements
    volatiles: Volatiles

    radio_tref: float = field(validator=gt(0))
    radio_U: float    = field(validator=ge(0))
    radio_K: float    = field(validator=ge(0))
    radio_Th: float   = field(validator=ge(0))



# Dummy interior module
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from proteus.interior.timestep import next_step
from proteus.utils.constants import secs_per_year

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def calculate_simple_mantle_mass(radius:float, corefrac:float)->float:
    '''
    A very simple interior structure model.

    This calculates mantle mass given planetary mass, radius, and core fraction. This
    assumes a core density equal to that of Earth's, and that the planet mass is simply
    the sum of mantle and core.
    '''

    # Volume of mantle shell
    mantle_volume = (4 * np.pi / 3) * (radius**3 - (radius*corefrac)**3)

    # Fixed density [g cm-3]
    # This is roughly representative of the average density across Earth's
    #    upper and lower mantle layers.
    mantle_rho = 4.55

    # Get mass [in SI units]
    mantle_mass = mantle_volume * mantle_rho * 1e3

    log.debug("Total mantle mass = %.2e kg" % mantle_mass)
    if mantle_mass <= 0.0:
        raise Exception("Something has gone wrong (mantle mass is negative)")

    return mantle_mass


# Run the dummy interior module
def RunDummyInt(config:Config, dirs:dict, IC_INTERIOR:int, hf_row:dict, hf_all:pd.DataFrame):

    # Output dictionary
    output = {}
    output["F_int"] = hf_row["F_atm"]

    # Interior structure
    output["M_mantle"] = calculate_simple_mantle_mass(hf_row["R_int"], config.struct.corefrac)

    # Parameters
    tmp_init = config.interior.dummy.ini_tmagma # Initial magma temperature
    tmp_liq  = 2700.0    # Liquidus
    tmp_sol  = 1700.0    # Solidus
    cp_m     = 1792.0    # Mantle specific heat capacity, J kg-1 K-1
    cp_c     = 880.0     # Core specific heat capacity, J kg-1 K-1
    area     = 4 * np.pi * hf_row["R_int"]**2

    # Get mantle melt fraction as a function of temperature
    def _calc_phi(tmp:float):
        # Too hot
        if tmp >= tmp_liq:
            return 1.0
        # Too cold
        elif tmp <= tmp_sol:
            return 0.0
        # Just right
        return  (tmp-tmp_sol)/(tmp_liq-tmp_sol )

    # Interior heat capacity [J K-1]
    cp_int = cp_m*output["M_mantle"] + cp_c*hf_row["M_core"]

    # Subtract tidal contribution to the total heat flux.
    #    This heat energy is generated only in the mantle, not in the core.
    tidal_flux = 0.0
    if config.interior.tidal_heat:
        if config.orbit.dummy:
            tidal_flux = config.orbit.dummy.H_tide * output["M_mantle"] / area
    output["F_tidal"] = tidal_flux

    # Radiogenic heating not included
    output["F_radio"] = 0.0

    # Total flux loss
    F_loss = output["F_int"] - output["F_tidal"] - output["F_radio"]

    # Rate of surface temperature change (this will be negative)
    dTdt = -F_loss * area / cp_int

    # Timestepping
    if IC_INTERIOR==1:
        output["T_magma"] = tmp_init
        dt = 0.0
    else:
        dt = next_step(config, dirs, hf_row, hf_all, 1.0)
        output["T_magma"] = hf_row["T_magma"] + dTdt * dt * secs_per_year

    # Determine the new melt fraction
    output["Phi_global"]        = _calc_phi(output["T_magma"])
    output["M_mantle_liquid"]   = output["M_mantle"] * output["Phi_global"]
    output["M_mantle_solid"]    = output["M_mantle"] - output["M_mantle_liquid"]
    output["RF_depth"]          = output["Phi_global"] * (1- config.struct.corefrac)

    sim_time = hf_row["Time"] + dt
    return sim_time, output



# Generic interior wrapper
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import scipy.optimize as optimise

from proteus.utils.constants import M_earth, R_earth, const_G, element_list
from proteus.utils.helper import UpdateStatusfile

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def update_gravity(hf_row:dict):
    '''
    Update surface gravity.
    '''
    hf_row["gravity"] = const_G * hf_row["M_int"] / (hf_row["R_int"] * hf_row["R_int"])

def calculate_core_mass(hf_row:dict, config:Config):
    '''
    Calculate the core mass of the planet.
    '''
    hf_row["M_core"] = config.struct.core_density * 4.0/3.0 * np.pi * (hf_row["R_int"] * config.struct.corefrac)**3.0

def determine_interior_radius(dirs:dict, config:Config, hf_all:pd.DataFrame, hf_row:dict):
    '''
    Determine the interior radius (R_int) of the planet.

    This uses the interior model's hydrostatic integration to estimate the planet's
    interior mass from a given radius. The radius is then adjusted until the interior mass
    achieves the target mass provided by the user in the config file.
    '''

    log.info("Using %s interior module to solve strcture"%config.interior.module)

    # Initial guess for interior radius and gravity
    IC_INTERIOR = 1
    hf_row["R_int"]   = config.struct.radius_int * R_earth
    calculate_core_mass(hf_row, config)
    hf_row["gravity"] = 9.81

    # Target mass
    M_target = config.struct.mass_tot * M_earth

    # We need to solve for the state hf_row[M_tot] = config.struct.mass_tot
    # This function takes R_int as the input value, and returns the mass residual
    def _resid(x):
        hf_row["R_int"] = x

        log.debug("Try R = %.2e m = %.3f R_earth"%(x,x/R_earth))

        # Use interior model to get dry mass from radius
        calculate_core_mass(hf_row, config)
        run_interior(dirs, config, IC_INTERIOR, hf_all, hf_row, verbose=False)
        update_gravity(hf_row)

        # Calculate residual
        res = hf_row["M_tot"] - M_target
        log.debug("    yields M = %.5e kg , resid = %.3e kg"%(hf_row["M_tot"], res))

        return res

    # Set tolerance
    match config.interior.module:
        case 'aragog':
            rtol = config.interior.aragog.tolerance
        case 'spider':
            rtol = config.interior.spider.tolerance
        case _:
            rtol = 1e-7

    # Find the radius
    r = optimise.root_scalar(_resid, method='secant', xtol=1e3, rtol=rtol, maxiter=10,
                                    x0=hf_row["R_int"], x1=hf_row["R_int"]*1.02)
    hf_row["R_int"] = float(r.root)
    calculate_core_mass(hf_row, config)
    run_interior(dirs, config, IC_INTERIOR, hf_all, hf_row)
    update_gravity(hf_row)

    # Result
    log.info("Found solution for interior structure")
    log.info("M_tot: %.1e kg = %.3f M_earth"%(hf_row["M_tot"], hf_row["M_tot"]/M_earth))
    log.info("R_int: %.1e m  = %.3f R_earth"%(hf_row["R_int"], hf_row["R_int"]/R_earth))
    log.info(" ")

def solve_structure(dirs:dict, config:Config, hf_all:pd.DataFrame, hf_row:dict):
    '''
    Solve for the planet structure based on the method set in the configuration file.

    If the structure is set by the radius, then this is trivial because the radius is used
    as an input to the interior modules anyway. If the structure is set by mass, then it is
    solved as an inverse problem for now.
    '''

    # Set total mass by radius
    # We might need here to setup a determine_interior_mass function as mass calculation depends on gravity
    if config.struct.set_by == 'radius_int':
        # radius defines interior structure
        hf_row["R_int"] = config.struct.radius_int * R_earth
        calculate_core_mass(hf_row, config)
        # initial guess for mass, which will be updated by the interior model
        hf_row["M_int"] = config.struct.mass_tot * M_earth
        update_gravity(hf_row)

    # Set by total mass (mantle + core + volatiles)
    elif config.struct.set_by == 'mass_tot':
        determine_interior_radius(dirs, config, hf_all, hf_row)

    # Otherwise, error
    else:
        log.error("Invalid constraint on interior structure: %s"%config.struct.set_by)


def run_interior(dirs:dict, config:Config, IC_INTERIOR:int,
                 hf_all:pd.DataFrame, hf_row:dict, verbose:bool=True):
    '''
    Run interior model
    '''

    # Use the appropriate interior model
    if verbose:
        log.info("Evolve interior...")
    log.debug("Using %s module to evolve interior"%config.interior.module)

    if config.interior.module == 'spider':
        # Import
        from proteus.interior.spider import ReadSPIDER, RunSPIDER

        # Run SPIDER
        RunSPIDER(dirs, config, IC_INTERIOR, hf_all, hf_row)
        sim_time, output = ReadSPIDER(dirs, config, hf_row["R_int"])

    elif config.interior.module == 'aragog':
        # Import
        from proteus.interior.aragog import RunAragog

        # Run Aragog
        sim_time, output = RunAragog(config, dirs, IC_INTERIOR, hf_row, hf_all)

    elif config.interior.module == 'dummy':
        # Import
        from proteus.interior.dummy import RunDummyInt

        # Run dummy interior
        sim_time, output = RunDummyInt(config, dirs, IC_INTERIOR, hf_row, hf_all)

    # Read output
    for k in output.keys():
        if k in hf_row.keys():
            hf_row[k] = output[k]

    # Ensure values are >= 0
    for k in ("M_mantle","M_mantle_liquid","M_mantle_solid","M_core","Phi_global"):
        hf_row[k] = max(hf_row[k], 0.0)

    # Check that the new temperature is remotely reasonable
    if not (0 < hf_row["T_magma"] < 1e6):
        log.error("T_magma is out of range: %g K"%hf_row["T_magma"])
        UpdateStatusfile(dirs, 21)
        exit(1)

    # Update dry interior mass
    hf_row["M_int"] = hf_row["M_mantle"]  + hf_row["M_core"]

    # Update total planet mass
    M_volatiles = 0.0
    for e in element_list:
        if (e == 'O'):
            # do not include oxygen, because it varies over time in order to set fO2.
            continue
        M_volatiles += hf_row[e+"_kg_total"]
    hf_row["M_tot"] = hf_row["M_int"] + M_volatiles

    # Prevent increasing melt fraction
    if config.atmos_clim.prevent_warming and (IC_INTERIOR==2):
        hf_row["Phi_global"] = min(hf_row["Phi_global"], hf_all.iloc[-1]["Phi_global"])
        hf_row["T_magma"] = min(hf_row["T_magma"], hf_all.iloc[-1]["T_magma"])

    # Print result of interior module
    if verbose:
        log.info("    T_magma    = %.3f K"%hf_row["T_magma"])
        log.info("    Phi_global = %.3f  "%hf_row["Phi_global"])
        log.info("    RF_depth   = %.2e W m-2" %hf_row["RF_depth"])
        log.info("    F_int      = %.2e W m-2" %hf_row["F_int"])
        log.info("    F_tidal    = %.2e W m-2" %hf_row["F_tidal"])
        log.info("    F_radio    = %.2e W m-2" %hf_row["F_radio"])


    # Time step size
    dt = float(sim_time) - hf_row["Time"]

    # Return timestep
    return dt

def get_all_output_times(output_dir:str, model:str):

    if model == "spider":
        from proteus.interior.spider import get_all_output_times as _get_output_times
    elif model == "aragog":
        from proteus.interior.aragog import get_all_output_times as _get_output_times
    else:
        return []

    return _get_output_times(output_dir)

def read_interior_data(output_dir:str, model:str, times:list):

    if len(times) == 0:
        return []

    if model == "spider":
        from proteus.interior.spider import read_jsons
        return read_jsons(output_dir, times)

    elif model == "aragog":
        from proteus.interior.aragog import read_ncdfs
        return read_ncdfs(output_dir, times)

    else:
        return []



from __future__ import annotations

import glob
import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import netCDF4 as nc
import numpy as np

from proteus.utils.helper import natural_sort

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def plot_fluxes_atmosphere(output_dir:str, plot_format="pdf"):

    log.info("Plot atmosphere fluxes")

    files = glob.glob(os.path.join(output_dir, "data", "*_atm.nc"))
    nc_fpath = natural_sort(files)[-1]

    # Read netCDF
    ds = nc.Dataset(nc_fpath)
    atm_pl      = np.array(ds.variables['pl'     ][:])
    atm_fl_D_LW = np.array(ds.variables['fl_D_LW'][:])
    atm_fl_U_LW = np.array(ds.variables['fl_U_LW'][:])
    atm_fl_D_SW = np.array(ds.variables['fl_D_SW'][:])
    atm_fl_U_SW = np.array(ds.variables['fl_U_SW'][:])
    atm_fl_D    = np.array(ds.variables['fl_D'   ][:])
    atm_fl_U    = np.array(ds.variables['fl_U'   ][:])
    atm_fl_N    = np.array(ds.variables['fl_N'   ][:])
    ds.close()

    scale = 1.0
    fig,ax = plt.subplots(1,1, figsize=(7*scale,6*scale ))

    ax.axvline(0,color='black',lw=0.8)

    pl = atm_pl * 1.e-5

    ax.plot(atm_fl_U      ,pl,color='red',label='UP',lw=1)
    ax.plot(atm_fl_U_SW   ,pl,color='red',label='UP SW',linestyle='dotted',lw=2)
    ax.plot(atm_fl_U_LW   ,pl,color='red',label='UP LW',linestyle='dashed',lw=1)

    ax.plot(-1.0*atm_fl_D      ,pl,color='green',label='DN',lw=2)
    ax.plot(-1.0*atm_fl_D_SW   ,pl,color='green',label='DN SW',linestyle='dotted',lw=3)
    ax.plot(-1.0*atm_fl_D_LW   ,pl,color='green',label='DN LW',linestyle='dashed',lw=2)

    ax.plot(atm_fl_N ,pl,color='black',label='NET')

    ax.set_xscale("symlog")
    ax.set_xlabel("Upward-directed flux [W m-2]")
    ax.set_ylabel("Pressure [bar]")
    ax.set_yscale("log")
    ax.set_ylim([pl[-1],pl[0]])

    ax.legend(loc='upper left')

    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_fluxes_atmosphere.%s"%plot_format,
                bbox_inches='tight', dpi=200)


def plot_fluxes_atmosphere_entry(handler: Proteus):
    plot_fluxes_atmosphere(
        output_dir=handler.directories["output"],
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == '__main__':
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_fluxes_atmosphere_entry(handler)



from __future__ import annotations

from proteus.plot.cpl_atmosphere import plot_atmosphere_entry
from proteus.plot.cpl_atmosphere_cbar import plot_atmosphere_cbar_entry
from proteus.plot.cpl_elements import plot_elements_entry
from proteus.plot.cpl_emission import plot_emission_entry
from proteus.plot.cpl_escape import plot_escape_entry
from proteus.plot.cpl_fluxes_atmosphere import plot_fluxes_atmosphere_entry
from proteus.plot.cpl_fluxes_global import plot_fluxes_global_entry
from proteus.plot.cpl_global import plot_global_entry
from proteus.plot.cpl_heatingrates import plot_heatingrates_entry
from proteus.plot.cpl_interior import plot_interior_entry
from proteus.plot.cpl_interior_cmesh import plot_interior_cmesh_entry
from proteus.plot.cpl_observables import plot_observables_entry
from proteus.plot.cpl_offchem_grid_cross import plot_offchem_grid_cross_entry
from proteus.plot.cpl_offchem_species import plot_offchem_species_entry
from proteus.plot.cpl_offchem_time import plot_offchem_time_entry
from proteus.plot.cpl_offchem_year import plot_offchem_year_entry
from proteus.plot.cpl_population import plot_population_entry
from proteus.plot.cpl_sflux import plot_sflux_entry
from proteus.plot.cpl_sflux_cross import plot_sflux_cross_entry
from proteus.plot.cpl_structure import plot_structure_entry

plot_dispatch = {
    'atmosphere':           plot_atmosphere_entry,
    'atmosphere_cbar':      plot_atmosphere_cbar_entry,
    'elements':             plot_elements_entry,
    'escape':               plot_escape_entry,
    'fluxes_atmosphere':    plot_fluxes_atmosphere_entry,
    'fluxes_global':        plot_fluxes_global_entry,
    'global':               plot_global_entry,
    'heatingrates':         plot_heatingrates_entry,
    'interior':             plot_interior_entry,
    'interior_cmesh':       plot_interior_cmesh_entry,
    'observables':          plot_observables_entry,
    'offchem_grid_cross':   plot_offchem_grid_cross_entry,
    'offchem_species':      plot_offchem_species_entry,
    'offchem_time':         plot_offchem_time_entry,
    'offchem_year':         plot_offchem_year_entry,
    'sflux':                plot_sflux_entry,
    'sflux_cross':          plot_sflux_cross_entry,
    'structure':            plot_structure_entry,
    'emission':             plot_emission_entry,
    'population':           plot_population_entry,
}



from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path

import numpy as np

from proteus.atmos_clim import RunAtmosphere
from proteus.config import read_config_object
from proteus.escape.wrapper import RunEscape
from proteus.interior.wrapper import run_interior, solve_structure
from proteus.orbit.wrapper import run_orbit
from proteus.outgas.wrapper import calc_target_elemental_inventories, run_outgassing
from proteus.star.wrapper import (
    get_new_spectrum,
    init_star,
    scale_spectrum_to_toa,
    update_stellar_mass,
    update_stellar_quantities,
    write_spectrum,
)
from proteus.utils.constants import (
    gas_list,
    vap_list,
    vol_list,
)
from proteus.utils.coupler import (
    CreateHelpfileFromDict,
    CreateLockFile,
    ExtendHelpfile,
    PrintCurrentState,
    ReadHelpfileFromCSV,
    SetDirectories,
    UpdatePlots,
    WriteHelpfileToCSV,
    ZeroHelpfileRow,
    print_header,
    print_module_configuration,
    print_system_configuration,
)
from proteus.utils.helper import (
    CleanDir,
    PrintHalfSeparator,
    PrintSeparator,
    UpdateStatusfile,
    safe_rm,
)
from proteus.utils.logs import (
    GetCurrentLogfileIndex,
    GetLogfilePath,
    setup_logger,
)
from proteus.utils.terminate import check_termination, print_termination_criteria


class Proteus:
    def __init__(self, *, config_path: Path | str) -> None:

        # Read and parse configuration file
        self.config_path = config_path
        self.config = read_config_object(config_path)

        # Setup directories dictionary
        self.init_directories()

        # Helpfile variables for the current iteration
        self.hf_row = None

        # Helpfile variables from all previous iterations
        self.hf_all = None

        # Loop counters
        self.loops = None

        # Interior
        self.IC_INTERIOR = -1   # initial condition (1: new simulation, 2: mid-simulation)
        self.dt = 1.0           # Time step variable [yr]

        # Model has finished?
        self.finished = False
        self.has_escaped = False
        self.lockfile = "/tmp/none"

        # Default values for mors.spada cases
        self.star_props  = None
        self.star_struct = None

        # Default values for mors.baraffe cases
        self.stellar_track  = None
        self.star_modern_fl = None
        self.star_modern_wl = None

        # Stellar spectrum (wavelengths, fluxes)
        self.star_wl = None
        self.star_fl = None

        # Time at which star was last updated
        self.sspec_prev = -np.inf   # spectrum
        self.sinst_prev = -np.inf   # instellation and radius

    def init_directories(self):
        """Initialize directories dictionary"""
        self.directories = SetDirectories(self.config)

    def start(self, *, resume: bool = False, offline: bool = False):
        """Start PROTEUS simulation.

        Parameters
        ----------
        resume : bool
            If True, continue from previous simulation.
        offline : bool
            Run in offline mode; do not try to connect to the internet.
        """

        # Import
        from proteus.utils.data import download_sufficient_data

        # First things
        start_time = datetime.now()
        self.config.params.resume = resume
        self.config.params.offline = offline
        UpdateStatusfile(self.directories, 0)

        # Clean output directory
        if not self.config.params.resume:
            CleanDir(self.directories["output"])
            CleanDir(os.path.join(self.directories["output"], "data"))

        # Get next logfile path
        logindex = 1 + GetCurrentLogfileIndex(self.directories["output"])
        logpath = GetLogfilePath(self.directories["output"], logindex)

        # Switch to logger
        setup_logger(logpath=logpath, logterm=True, level=self.config.params.out.logging)
        log = logging.getLogger("fwl."+__name__)

        # Print header
        print_header()

        # Print system configuration
        print_system_configuration(self.directories)

        # Print module configuration
        print_module_configuration(self.directories, self.config, self.config_path)

        # Print termination criteria
        print_termination_criteria(self.config)

        PrintHalfSeparator()

        # Count iterations
        self.loops = {
            "total": 0,  # Total number of iters performed
            "total_min": self.config.params.stop.iters.minimum,
            "total_loops": self.config.params.stop.iters.maximum,
            "init": 0,  # Number of init iters performed
            "init_loops": 3,  # Maximum number of init iters
        }

        # Write config to output directory, for future reference
        self.config.write(os.path.join(self.directories["output"], "init_coupler.toml"))

        # Create lockfile for keeping simulation running
        self.lockfile = CreateLockFile(self.directories["output"])

        # Download basic data
        download_sufficient_data(self.config)

        # Is the model resuming from a previous state?
        if not self.config.params.resume:
            # New simulation

            # SPIDER initial condition
            self.IC_INTERIOR = 1

            # Create an empty initial row for helpfile
            self.hf_row = ZeroHelpfileRow()

            # Stellar mass
            update_stellar_mass(self.hf_row, self.config)

            # Initial time
            self.hf_row["Time"] = 0.0
            self.hf_row["age_star"] = self.config.star.age_ini * 1e9

            # Initial guess for flux
            self.hf_row["F_atm"] = self.config.interior.F_initial
            self.hf_row["F_int"] = self.hf_row["F_atm"]
            self.hf_row["T_eqm"] = 2000.0

            # Solve interior structure
            solve_structure(self.directories, self.config, self.hf_all, self.hf_row)

            # Store partial pressures and list of included volatiles
            inc_gases = []
            for s in vol_list:
                pp_val = self.config.delivery.volatiles.get_pressure(s)
                include = self.config.outgas.calliope.is_included(s)

                if include:
                    inc_gases.append(s)
                    self.hf_row[s + "_bar"] = max(1.0e-30, float(pp_val))
                else:
                    self.hf_row[s + "_bar"] = 0.0
            for s in vap_list:
                inc_gases.append(s)
                self.hf_row[s + "_bar"] = 0.0

            # Inform user
            log.info("Initial inventory set by '%s'"%self.config.delivery.initial)
            log.info("Included gases")
            for s in inc_gases:
                log.info("    %s  %-8s : %6.2f bar" %
                            ("vapour  " if s in vap_list else "volatile", s,
                            self.hf_row[s + "_bar"])
                        )
        else:
            # Resuming from disk
            log.info("Resuming the simulation from the disk")

            # SPIDER initial condition
            self.IC_INTERIOR = 2

            # Read helpfile from disk
            self.hf_all = ReadHelpfileFromCSV(self.directories["output"])

            # Check length
            if len(self.hf_all) <= self.loops["init_loops"] + 1:
                raise Exception("Simulation is too short to be resumed")

            # Get last row
            self.hf_row = self.hf_all.iloc[-1].to_dict()

            # Set loop counters
            self.loops["total"] = len(self.hf_all)
            self.loops["init"] = self.loops["init_loops"] + 1
        log.info(" ")

        # Prepare star stuff
        init_star(self)

        # Main loop
        UpdateStatusfile(self.directories, 1)
        while not self.finished:
            # New rows
            if self.loops["total"] > 0:
                # Create new row to hold the updated variables. This will be
                #    overwritten by the routines below.
                self.hf_row = self.hf_all.iloc[-1].to_dict()

            log.info(" ")
            PrintSeparator()
            log.info("Loop counters")
            log.info("init      total")
            log.info(
                "%1d/%1d     %04d/%04d "
                % (
                    self.loops["init"],
                    self.loops["init_loops"],
                    self.loops["total"],
                    self.loops["total_loops"],
                )
            )

            ############### INTERIOR

            PrintHalfSeparator()

            # Run interior model
            self.dt = run_interior(self.directories, self.config,
                                    self.IC_INTERIOR, self.hf_all,  self.hf_row)


            # Advance current time in main loop according to interior step
            self.hf_row["Time"]     += self.dt    # in years
            self.hf_row["age_star"] += self.dt    # in years

            ############### / INTERIOR AND STRUCTURE

            ############### ORBIT AND TIDES

            PrintHalfSeparator()
            run_orbit(self.hf_row, self.config)

            ############### / ORBIT AND TIDES

            ############### STELLAR FLUX MANAGEMENT
            PrintHalfSeparator()
            log.info("Stellar flux management...")
            update_stellar_spectrum = False

            # Calculate new instellation and radius
            if (abs(self.hf_row["Time"] - self.sinst_prev) > self.config.params.dt.starinst) or (
                self.loops["total"] == 0
            ):
                self.sinst_prev = self.hf_row["Time"]

                update_stellar_quantities(self.hf_row, self.config, stellar_track=self.stellar_track)

            # Calculate a new (historical) stellar spectrum
            if (abs(self.hf_row["Time"] - self.sspec_prev) > self.config.params.dt.starspec) or (
                self.loops["total"] == 0
            ):
                self.sspec_prev = self.hf_row["Time"]
                update_stellar_spectrum = True

                # Get the new spectrum using the appropriate module
                log.info("Updating stellar spectrum")
                self.star_wl, self.star_fl = get_new_spectrum(

                                        # Required variables
                                        self.hf_row["age_star"], self.config,

                                        # Variables needed for mors.spada
                                        star_struct_modern=self.star_struct,
                                        star_props_modern=self.star_props,

                                        # Variables needed for mors.baraffe
                                        stellar_track=self.stellar_track,
                                        modern_wl=self.star_modern_wl,
                                        modern_fl=self.star_modern_fl,
                                        )

                # Scale fluxes from 1 AU to TOA
                self.star_fl = scale_spectrum_to_toa(self.star_fl, self.hf_row["separation"])

                # Save spectrum to file
                write_spectrum(self.star_wl, self.star_fl,
                                self.hf_row, self.directories["output"])

            else:
                log.info("New spectrum not required at this time")

            ############### / STELLAR FLUX MANAGEMENT

            ############### ESCAPE
            if (self.loops["total"] >= self.loops["init_loops"]):
                RunEscape(self.config, self.hf_row, self.dt, self.stellar_track)

            ############### / ESCAPE

            ############### OUTGASSING
            PrintHalfSeparator()

            #    recalculate mass targets during init phase, since these will be adjusted
            #    depending on the true melt fraction and T_magma found by SPIDER at runtime.
            if self.loops["init"] < self.loops["init_loops"]:
                calc_target_elemental_inventories(self.directories, self.config, self.hf_row)

            # solve for atmosphere composition
            run_outgassing(self.directories, self.config, self.hf_row)

            # Add atmosphere mass to interior mass, to get total planet mass
            self.hf_row["M_planet"] = self.hf_row["M_int"] + self.hf_row["M_atm"]

            # Check for when atmosphere has escaped.
            #    This will mean that the mixing ratios become undefined, so use value of 0.
            if self.hf_row["P_surf"] < 1.0e-10:
                self.has_escaped = True
                for gas in gas_list:
                    self.hf_row[gas+"_vmr"] = 0.0
                self.hf_row["atm_kg_per_mol"] = 0.0

            ############### / OUTGASSING

            ############### ATMOSPHERE CLIMATE
            if not self.has_escaped:
                RunAtmosphere(self.config, self.directories, self.loops,
                                self.star_wl, self.star_fl, update_stellar_spectrum,
                                self.hf_all, self.hf_row)

            ############### HOUSEKEEPING AND CONVERGENCE CHECK

            PrintHalfSeparator()

            # Update init loop counter
            # Next init iter
            if self.loops["init"] < self.loops["init_loops"]:
                self.loops["init"] += 1
                self.hf_row["Time"] = 0.0
            # Reset restart flag once SPIDER has correct heat flux
            if self.loops["total"] >= self.loops["init_loops"]:
                self.IC_INTERIOR = 2

            # Adjust total iteration counters
            self.loops["total"] += 1

            # Update full helpfile
            if self.loops["total"] > 1:
                # append row
                self.hf_all = ExtendHelpfile(self.hf_all, self.hf_row)
            else:
                # first iter => generate new HF from dict
                self.hf_all = CreateHelpfileFromDict(self.hf_row)

            # Write helpfile to disk
            WriteHelpfileToCSV(self.directories["output"], self.hf_all)

            # Print info to terminal and log file
            PrintCurrentState(self.hf_row)

            # Check for convergence
            if self.loops["total"] >= self.loops["init_loops"]:
                log.info("Check convergence criteria")
                self.finished = check_termination(self)

            # Make plots
            if self.config.params.out.plot_mod > 0 and not self.finished:
                if self.loops["total"] % self.config.params.out.plot_mod == 0:
                    log.info("Making plots")
                    UpdatePlots(self.hf_all, self.directories, self.config)

            ############### / HOUSEKEEPING AND CONVERGENCE CHECK

        # FINAL THINGS BEFORE EXIT
        PrintSeparator()

        # Clean up files
        safe_rm(self.lockfile)

        # Plot conditions at the end
        UpdatePlots(self.hf_all, self.directories, self.config, end=True)
        end_time = datetime.now()
        log.info("Simulation stopped at: " + end_time.strftime("%Y-%m-%d_%H:%M:%S"))

        run_time = end_time - start_time
        run_time = run_time.total_seconds() / 60  # minutes
        if run_time > 60:
            log.info("Total runtime: %.2f hours" % (run_time / 60))
        else:
            log.info("Total runtime: %.2f minutes" % run_time)



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm
from matplotlib.ticker import LogLocator

from proteus.atmos_clim.common import read_atmosphere_data
from proteus.utils.plot import latex_float, sample_output

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def plot_atmosphere( output_dir:str, times:list, profiles:list, plot_format="pdf"):

    if np.amax(times) < 2:
        log.debug("Insufficient data to make plot_atmosphere")
        return

    log.info("Plot atmosphere temperatures")

    norm = mpl.colors.LogNorm(vmin=max(times[0],1), vmax=times[-1])
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])

    scale = 1.2
    fig,(ax0,ax1) = plt.subplots(2,1, sharex=True, figsize=(5*scale,7*scale))

    for i, t in enumerate( times ):
        prof = profiles[i]
        label = latex_float(t)+" yr"
        color = sm.to_rgba(t)
        ax0.plot( prof["t"], prof["z"]/1e3, color=color, label=label, lw=1.5)
        ax1.plot( prof["t"], prof["p"]/1e5, color=color, label=label, lw=1.5)

    #####  T-Z
    # fig_o.set_myaxes( ax0, xlabel='$T$ (K)', ylabel='$z_\mathrm{atm}$\n(km)', xmin=xmin, xmax=xmax, ymin=0, ymax=ymax_atm_z, xticks=xticks )
    ax0.set_ylabel(r"Height [km]")

    #####  T-P
    # fig_o.set_myaxes( ax1, xlabel='$T$ (K)', ylabel='$P_\mathrm{atm}$\n(bar)', xmin=xmin, xmax=xmax, xticks=xticks )
    ax1.set_xlabel("Temperature [K]")
    ax1.set_ylabel("Pressure [bar]")
    ax1.invert_yaxis()
    ax1.set_yscale("log")

    # Legend
    ax1.legend( fontsize=8, fancybox=True, framealpha=0.5 )
    ax1.yaxis.set_major_locator(LogLocator(numticks=1000))

    fig.subplots_adjust(hspace=0.02)

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_atmosphere.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_atmosphere_entry(handler: Proteus):
    plot_times, _ = sample_output(handler, tmin=1e4)
    print("Snapshots:", plot_times)

    # Plot fixed set from above
    profiles = read_atmosphere_data(handler.directories["output"], plot_times)
    plot_atmosphere(
        output_dir=handler.directories["output"],
        times=plot_times,
        profiles=profiles,
        plot_format=handler.config.params.out.plot_fmt,
   )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_atmosphere_entry(handler)



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from proteus.utils.constants import M_earth, s2yr

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def plot_escape(hf_all:pd.DataFrame, output_dir:str, escape_model, plot_format="pdf", t0=100.0) :

    time = np.array(hf_all["Time"])
    if np.amax(time) < 2:
        log.debug("Insufficient data to make plot_escape")
        return

    log.info("Plot escape")

    # make plot
    lw = 1.2
    scale = 1.1
    fig,ax1 = plt.subplots(1,1, figsize=(7*scale,4*scale))

    if not escape_model:
        escape_model_label = 'No escape'
    elif escape_model == 'zephyrus':
        escape_model_label = 'Energy-limited escape (Zephyrus)'
    elif escape_model == 'dummy':
        escape_model_label = 'Dummy escape'

    y = hf_all['esc_rate_total']
    ax1.plot(time, y, lw=lw, ls='solid', label=f'{escape_model_label}')

    # decorate
    ax1.set_ylabel(r'Mass loss rate [kg $s^{-1}$]')
    ax1.set_yscale("log")
    ax1.set_xlabel("Time [yr]")
    ax1.set_xscale("log")
    ax1.set_xlim(left=t0, right=np.amax(time))
    ax1.grid(alpha=0.2)
    ax1.legend()
    ax2 = ax1.twinx()
    ylims = ax1.get_ylim()
    ax2.set_ylim((ylims[0]/ s2yr) / M_earth,(ylims[1] / s2yr) / M_earth)
    ax2.set_yscale('log')
    ax2.set_ylabel(r'Mass loss rate [$M_{\oplus}$ $yr^{-1}$]')

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_escape.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_escape_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plot
    plot_escape(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        escape_model=handler.config.escape.module,
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_escape_entry(handler)



from __future__ import annotations

import functools
import logging
import os
import subprocess as sp
from pathlib import Path
from typing import TYPE_CHECKING

import platformdirs
from osfclient.api import OSF

if TYPE_CHECKING:
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

FWL_DATA_DIR = Path(os.environ.get('FWL_DATA', platformdirs.user_data_dir('fwl_data')))

log.debug(f'FWL data location: {FWL_DATA_DIR}')

def download_folder(*, storage, folders: list[str], data_dir: Path):
    """
    Download a specific folder in the OSF repository

    Inputs :
        - storage : OSF storage name
        - folders : folder names to download
        - data_dir : local repository where data are saved
    """
    for file in storage.files:
        for folder in folders:
            if not file.path[1:].startswith(folder):
                continue
            parts = file.path.split('/')[1:]
            target = Path(data_dir, *parts)
            target.parent.mkdir(parents=True, exist_ok=True)
            log.info(f'Downloading {file.path}...')
            with open(target, 'wb') as f:
                file.write_to(f)
            break

def GetFWLData() -> Path:
    """
    Get path to FWL data directory on the disk
    """
    return Path(FWL_DATA_DIR).absolute()


@functools.cache
def get_osf(id: str):
    """
    Generate an object to access OSF storage
    """
    osf = OSF()
    project = osf.project(id)
    return project.storage('osfstorage')


def download(
    *,
    folder: str,
    target: str,
    osf_id: str,
    desc: str

):
    """
    Generic download function.

    Attributes
    ----------
    folder: str
        Filename to download
    target: str
        name of target directory
    osf_id: str
        OSF project id
    desc: str
        Description for logging
    """
    log.debug(f"Get {desc}?")

    data_dir = GetFWLData() / target
    data_dir.mkdir(parents=True, exist_ok=True)

    if not (data_dir / folder).exists():
        storage = get_osf(osf_id)
        log.info(f"Downloading {desc} to {data_dir}")
        download_folder(storage=storage, folders=[folder], data_dir=data_dir)
    else:
        log.debug(f"\t{desc} already exists")


def download_surface_albedos():
    """
    Download surface optical properties
    """
    download(
        folder = 'Hammond24',
        target = "surface_albedos",
        osf_id = '2gcd9',
        desc = 'surface albedos'
    )


def download_spectral_file(name:str, bands:str):
    """
    Download spectral file.

    Inputs :
        - name : str
            folder name (e.g. "Dayspring")
        - bands : str
            number of bands (e.g. "256")
    """
    # Check name and bands
    if not isinstance(name, str) or (len(name) < 1):
        raise Exception("Must provide name of spectral file")
    if not isinstance(bands, str) or (len(bands) < 1):
        raise Exception("Must provide number of bands in spectral file")

    download(
        folder = f'{name}/{bands}',
        target = "spectral_files",
        osf_id = 'vehxg',
        desc = f'{name}{bands} spectral file',
    )


def download_stellar_spectra():
    """
    Download stellar spectra
    """
    download(
        folder = 'Named',
        target = "stellar_spectra",
        osf_id = '8r2sw',
        desc = 'stellar spectra'
    )


def download_exoplanet_data():
    """
    Download exoplanet data
    """
    download(
        folder = 'Exoplanets',
        target = "planet_reference",
        osf_id = 'fzwr4',
        desc = 'exoplanet data'
    )


def download_massradius_data():
    """
    Download mass-radius data
    """
    download(
        folder = 'Mass-radius',
        target = "mass_radius",
        osf_id = 'fzwr4',
        desc = 'mass radius data'
    )


def download_evolution_tracks(track:str):
    """
    Download evolution tracks
    """
    from mors.data import DownloadEvolutionTracks
    log.debug("Get evolution tracks")
    DownloadEvolutionTracks(track)

def download_interior_lookuptables():
    """
    Download interior lookup tables
    """
    from aragog.data import DownloadLookupTableData
    log.debug("Get interior lookup tables")
    DownloadLookupTableData()


def _get_sufficient(config:Config):
    # Star stuff
    if config.star.module == "mors":
        download_stellar_spectra()
        if config.star.mors.tracks == 'spada':
            download_evolution_tracks("Spada")
        else:
            download_evolution_tracks("Baraffe")

    # Spectral files
    if config.atmos_clim.module in ('janus', 'agni'):
        # High-res file often used for post-processing
        download_spectral_file("Honeyside","4096")

        # Get the spectral file we need for this simluation
        from proteus.atmos_clim.common import get_spfile_name_and_bands
        group, bands = get_spfile_name_and_bands(config)
        download_spectral_file(group, bands)


    # Surface single-scattering data
    if config.atmos_clim.module == 'agni':
        download_surface_albedos()

    # Exoplanet population data
    download_exoplanet_data()

    # Mass-radius reference data
    download_massradius_data()

    # Interior look up tables
    if config.interior.module == "aragog":
        download_interior_lookuptables()


def download_sufficient_data(config:Config):
    """
    Download the required data based on the current options
    """

    log.info("Getting physical and reference data")

    if config.params.offline:
        # Don't try to get data
        log.warning("Running in offline mode. Will not check for reference data.")

    else:
        # Try to get data
        try:
            _get_sufficient(config)

        # Some issue. Usually due to lack of internet connection, but print the error
        #     anyway so that the user knows what happened.
        except OSError as e:
            log.warning("Problem when downloading/checking reference data")
            log.warning(str(e))

    log.info(" ")


def _none_dirs():
    from proteus.utils.helper import get_proteus_dir

    dirs = {"proteus":get_proteus_dir()}
    dirs["tools"] = os.path.join(dirs["proteus"],"tools")
    return dirs


def get_socrates(dirs=None):
    """
    Download and install SOCRATES
    """

    log.info("Setting up SOCRATES")

    # None dirs
    if dirs is None:
        dirs = _none_dirs()

    # Get path
    workpath = os.path.join(dirs["proteus"], "SOCRATES")
    workpath = os.path.abspath(workpath)
    if os.path.isdir(workpath):
        log.debug("\talready set up")
        return

    # Download, configure, and build
    log.debug("Running get_socrates.sh")
    cmd = [os.path.join(dirs["tools"],"get_socrates.sh"), workpath]
    out = os.path.join(dirs["proteus"], "nogit_setup_socrates.log")
    log.debug("    logging to %s"%out)
    with open(out,'w') as hdl:
        sp.run(cmd, check=True, stdout=hdl, stderr=hdl)

    # Set environment
    os.environ["RAD_DIR"] = workpath
    log.debug("\tdone")


def get_petsc(dirs=None):
    """
    Download and install PETSc
    """

    log.info("Setting up PETSc")

    if dirs is None:
        dirs = _none_dirs()

    # Get path
    workpath = os.path.join(dirs["proteus"], "petsc")
    workpath = os.path.abspath(workpath)
    if os.path.isdir(workpath):
        # already downloaded
        log.debug("\talready set up")
        return

    # Download, configure, and build
    log.debug("Running get_petsc.sh")
    cmd = [os.path.join(dirs["tools"],"get_petsc.sh"), workpath]
    out = os.path.join(dirs["proteus"], "nogit_setup_petsc.log")
    log.debug("\tlogging to %s"%out)
    with open(out,'w') as hdl:
        sp.run(cmd, check=True, stdout=hdl, stderr=hdl)

    log.debug("\tdone")


def get_spider(dirs=None):
    """
    Download and install SPIDER
    """

    if dirs is None:
        dirs = _none_dirs()

    # Need to install PETSc first
    get_petsc(dirs)

    log.info("Setting up SPIDER")

    # Get path
    workpath = os.path.join(dirs["proteus"], "SPIDER")
    workpath = os.path.abspath(workpath)
    if os.path.isdir(workpath):
        # already downloaded
        log.debug("\talready set up")
        return

    # Download, configure, and build
    log.debug("Running get_spider.sh")
    cmd = [os.path.join(dirs["tools"],"get_spider.sh"), workpath]
    out = os.path.join(dirs["proteus"], "nogit_setup_spider.log")
    log.debug("\tlogging to %s"%out)
    with open(out,'w') as hdl:
        sp.run(cmd, check=True, stdout=hdl, stderr=hdl)

    log.debug("\tdone")



from __future__ import annotations

import glob
import logging
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm

from proteus.utils.constants import const_c, const_h, const_k
from proteus.utils.helper import find_nearest, natural_sort

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def planck_function(lam, T):
    """Plots stellar flux from output directory for a set of wavelength bins

    Planck function value at stellar surface
    lam in nm
    erg s-1 cm-2 nm-1
    """

    x = lam * 1.0e-9   # convert nm -> m
    hc_by_kT = const_h*const_c / (const_k*T)

    planck_func = 1.0/( (x ** 5.0) * ( np.exp( hc_by_kT/ x) - 1.0 ) )
    planck_func *= 2 * const_h * const_c * const_c #  w m-2 sr-1 s-1 m-1
    planck_func *= np.pi * 1.0e3 * 1.0e-9  # erg s-1 cm-2 nm-1

    return planck_func


def plot_sflux_cross(
        output_dir: str,
        wl_targets: list | None=None,
        modern_age: float=-1,
        plot_format="pdf",
    ):
    """Plots stellar flux vs time, for a set of wavelengths.

    Note that this function will plot the flux from EVERY file it finds.
    Saves plot as 'cpl_sflux_cross.pdf' in  the output directory.

    Parameters
    ----------
    output_dir : str
        Directory for both reading from and saving to.
    wl_targets : list | None
        List of wavelengths to plot [nm]
    modern_age : float
        Current age of star. If not provided, then won't be plotted
    """

    mpl.use('Agg')

    # Wavelength targets default value
    if not wl_targets:
        wl_targets = [1.0, 12.0, 50.0, 121.0, 200.0, 400.0, 500.0, 2000.0]

    # Find and sort files
    files = glob.glob(output_dir+"/data/*.sflux")
    files = natural_sort(files)

    if (len(files) == 0):
        log.warning("No files found when trying to make sflux_cross plot")
        return
    if (len(files) == 1):
        log.warning("Cannot make sflux_cross plot with only 1 stellar spectrum sample")
        return

    log.info("Plot stellar flux (crossection)")

    # Arrays for storing data over time
    time_t = []
    wave_t = []
    flux_t = []
    for f in files:
        # Load data
        X = np.loadtxt(f,skiprows=1,delimiter='\t').T

        # Parse data
        time = int(f.split('/')[-1].split('.')[0])
        if (time < 0):
            continue
        wave = X[0]
        flux = X[1]

        # Save data
        time_t.append(time)
        wave_t.append(wave)
        flux_t.append(flux)

    time_t = np.array(time_t)
    wave_t = np.array(wave_t)
    flux_t = np.array(flux_t)

    # Create figure
    fig,ax = plt.subplots(1,1)

    ax.set_yscale("log")
    ax.set_ylabel("Flux [erg / (s cm$^2$ nm)]")

    ax.set_xscale("log")
    ax.set_xlabel("Time [yr]")
    ax.set_title("TOA flux versus time")

    vmin = max(time_t[0],1.0)
    vmax = time_t[-1]
    ax.set_xlim([vmin,vmax])

    # Find indices for wavelength bins
    wl_iarr = []
    wl_varr = []
    for w in wl_targets:
        wl_v, wl_i = find_nearest(wave,w)
        wl_iarr.append(wl_i)
        wl_varr.append(wl_v)
    N = len(wl_iarr)

    # Load modern spectrum
    if modern_age > 0:
        X = np.loadtxt(output_dir+'/-1.sflux',skiprows=2).T

    # Plot bins over time
    for i in range(N):

        fl = flux_t.T[wl_iarr[i]]
        c = cm.oleron(1.0*i/N)
        lbl = "%d"%max(1,round(wl_varr[i]))

        ax.plot(time_t,fl,color=c,lw=2.8,label=lbl)

        # Plot modern values
        if modern_age > 0:
            ax.scatter(modern_age,X[1][wl_iarr[i]],marker='o',color=c, s=40, zorder=4, edgecolors='white')

    leg = ax.legend(title=r"$\lambda$ [nm]", loc='center left',bbox_to_anchor=(1.02, 0.5))
    for legobj in leg.legend_handles:
        legobj.set_linewidth(4.0)

    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_sflux_cross.%s"%plot_format,
                bbox_inches="tight", dpi=200)


def plot_sflux_cross_entry(handler: Proteus):
    wl_targets = [1.0, 12.0, 50.0, 121.0, 200.0, 400.0, 500.0, 2000.0]

    if handler.config.star.module == "mors":
        modern_age = handler.config.star.age_now * 1e9
    else:
        modern_age = -1

    plot_sflux_cross(
        output_dir=handler.directories['output'],
        wl_targets=wl_targets,
        modern_age=modern_age,
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == '__main__':
    print("Plotting stellar flux over time (bins)...")

    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_sflux_cross_entry(handler)

    print("Done!")



# Small helper functions that can be used universally
# This file should not depend on too many other files, as this can cause circular import issues
from __future__ import annotations

import glob
import logging
import os
import re
import shutil

import numpy as np

log = logging.getLogger("fwl."+__name__)

def get_proteus_dir():
    '''
    Get absolute path to PROTEUS directory.

    This should be a directory containing `pyproject.toml`.
    '''

    # Assuming that this file is in `PROTEUS/src/proteus/utils/`
    utils = os.path.dirname(os.path.abspath(__file__))

    # Work upwards from utils
    root = os.path.abspath(os.path.join(utils,"..","..",".."))

    # Check that this path is reasonable
    if "pyproject.toml" not in os.listdir(root):
        raise EnvironmentError(f"Cannot locate PROTEUS directory. Tried '{root}' ")

    return root

def PrintSeparator():
    log.info("===================================================")
    pass

def PrintHalfSeparator():
    log.info("---------------------------------------------------")
    pass

def mol_to_ele(mol:str):
    '''
    Return the number of atoms of each element in a given molecule, as a dictionary

    https://codereview.stackexchange.com/a/232664
    '''

    # Validate
    if not str(mol[0]).isupper:
        raise ValueError(f"Molecule name '{mol}' is invalid")

    # Get atoms
    decomp = re.findall(r'([A-Z][a-z]?)(\d*)', mol)
    elems = {}
    for ev in decomp:
        if ev[1] == '':
            val = 1
        else:
            val = int(ev[1])
        elems[str(ev[0])] = val

    # Check that what we got is reasonable
    if not elems:
        raise ValueError(f"Could not decompose molecule '{mol}'")

    return elems

# String sorting inspired by natsorted
def natural_sort(lst):
    def convert(text):
        return int(text) if text.isdigit() else text.lower()
    def alphanum_key(key):
        return [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(lst, key = alphanum_key)

# Create a temporary folder
def create_tmp_folder():
    tmp_dir = "/tmp/proteus_%d/" % np.random.randint(int(1e12),int(1e13-1))
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir,ignore_errors=True)
    os.makedirs(tmp_dir)
    return tmp_dir

def safe_rm(fpath:str):
    '''
    Safely remove a file or folder
    '''

    if fpath=="":
        log.warning("Could not remove file at empty path")
        return

    fpath = os.path.abspath(fpath)
    if os.path.exists(fpath):
        if os.path.isfile(fpath):
            os.remove(fpath)

        elif os.path.isdir(fpath):
            subfolders = [ f.path.split("/")[-1] for f in os.scandir(fpath) if f.is_dir() ]
            if ".git" in subfolders:
                log.warning("Not emptying directory '%s' as it contains a Git repository"%fpath)
                return
            shutil.rmtree(fpath)

        else:
            log.warning("Cannot remove unhandled path '%s'"%fpath)

def CommentFromStatus(status:int):
    '''
    Convert status number into comment string
    '''
    desc = ""
    match status:
        # Running cases
        case 0:
            desc = "Started"
        case 1:
            desc = "Running"
        # Successful cases
        case 10:
            desc = "Completed (solidified)"
        case 11:
            desc = "Completed (steady-state)"
        case 12:
            desc = "Completed (maximum iterations)"
        case 13:
            desc = "Completed (target time)"
        case 14:
            desc = "Completed (net flux is small)"
        case 15:
            desc = "Completed (atmosphere escaped)"
        # Error cases
        case 20:
            desc = "Error (generic case, or configuration issue)"
        case 21:
            desc = "Error (Interior model)"
        case 22:
            desc = "Error (Atmosphere model)"
        case 23:
            desc = "Error (Stellar evolution model)"
        case 24:
            desc = "Error (Kinetics model)"
        case 25:
            desc = "Error (died, or exit requested by user)"
        # Default case
        case _:
            desc = "UNHANDLED STATUS (%d)" % status
            log.warning("Unhandled model status (%d) selected" % status)
    return desc

def UpdateStatusfile(dirs:dict, status:int):
    '''
    Update the status file with the current state of the program
    '''

    # Path to status file
    stsfold = os.path.abspath(dirs["output"])
    stsfile = os.path.join(stsfold,"status")

    # Does the folder exist?
    if not os.path.exists(stsfold):
        os.makedirs(stsfold)

    # Does the status file exist?
    safe_rm(stsfile)

    # Write status file
    with open(stsfile,'x') as hdl:
        hdl.write("%d\n" % status)
        desc = CommentFromStatus(status)
        hdl.write("%s\n"%desc)

def CleanDir(directory, keep_stdlog=False):
    """Clean a directory.

    Deletes a given directory and its contents, then creates it as empty.

    Parameters
    ----------
        directory : string
            Path to directory

    """

    def _check_safe(d):
        subfolders = [ f.path.split("/")[-1] for f in os.scandir(d) if f.is_dir() ]
        if ".git" in subfolders:
            raise Exception("Not emptying directory - it contains a Git repository!")

    # Simple case...
    if not keep_stdlog:
        if os.path.exists(directory):
            _check_safe(directory)
            shutil.rmtree(directory)
        os.makedirs(directory)
        return

    # Case where we want to keep log file...
    # If exists
    if os.path.exists(directory):
        for p in glob.glob(directory+"/*"):
            p = str(p)
            if os.path.isdir(p):
                # Remove folders
                _check_safe(p)
                shutil.rmtree(p)
            else:
                # Remove all files EXCEPT logfiles in topmost dir
                if ".log" not in p:
                    os.remove(p)
    else:
        os.makedirs(directory)


def find_nearest(array, target):
    """Find the element of an array that has a value closest to the target

    Parameters
    ----------
        array : list
            Array to search
        target : float
            Value to approximate

    Returns
    ----------
        close : float
            Element of array that is closest to target
        idx : int
            Index of closest element of array
    """
    array   = np.asarray(array)
    idx     = (np.abs(array - target)).argmin()
    close   = array[idx]
    return close, idx

def recursive_get(d, keys):
    '''
    Function to access nested dictionaries
    '''
    if len(keys) == 1:
        return d[keys[0]]
    return recursive_get(d[keys[0]], keys[1:])



from __future__ import annotations

import glob
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from cmcrameri import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable

from proteus.utils.plot import latexify
from proteus.utils.plot_offchem import offchem_read_year

if TYPE_CHECKING:
    from proteus import Proteus

mpl.use("Agg")


def plot_offchem_species(
        output_dir: str,
        specie: str,
        tmin: float=-1.0,
        tmax: float=-1.0,
        plot_init_mx: bool=False
    ):
    """Plot evolution of a single component according to offline VULCAN output

    Reads-in the data from output_dir for a single species. Can also
    include JANUS/SPIDER mixing ratios as dashed lines, which were used to
    initialise each VULCAN run.

    Parameters
    ----------
    output_dir : str
        Output directory that was specified in the PROTEUS cfg file
    specie : str
        Which species to plot? (e.g. H2O)
    tmin : float
        Initial year to include (-1 for start of data)
    tmax : float
        Final year to include (-1 for end of data)
    plot_init_mx : bool
        Include initial mixing ratios for each VULCAN run in plot?
    """

    mpl.use("Agg")

    ls = glob.glob(output_dir+"/offchem/*/output.vul")
    years = [int(f.split("/")[-2]) for f in ls]
    years_data = [offchem_read_year(output_dir,y,read_const=plot_init_mx) for y in years]

    if len(years) == 0:
        raise Exception('No VULCAN output files found')

    lw = 1.3
    alpha = 0.7

    fig, (ax0,ax1) = plt.subplots(1,2,sharey=True,figsize=(8,4))

    ax0.set_ylabel("Pressure [bar]")
    ax0.invert_yaxis()
    ax0.set_yscale("log")
    ax0.set_xlabel("Temperature [K]")
    ax0.set_title("Evolution of $T(p)$")

    ax1.set_title("Evolution of "+latexify(specie))
    ax1.set_xlabel("Mole fraction")
    ax1.set_xscale("log")

    divider = make_axes_locatable(ax1)
    cax = divider.append_axes('right', size='5%', pad=0.05)

    if (tmin > 0):
        cb_vmin = tmin
    else:
        cb_vmin = max(years_data[1]["year"],1.0)


    if (tmax > 0):
        cb_vmax = tmax
    else:
        cb_vmax = max(years_data[-1]["year"],2.0)


    norm = mpl.colors.LogNorm(vmin=cb_vmin, vmax=cb_vmax)
    sm = plt.cm.ScalarMappable(cmap=cm.batlowK_r, norm=norm)
    sm.set_array([])
    cbar = fig.colorbar(sm, cax=cax, orientation='vertical')  #, ticks=np.linspace(0,2,N), boundaries=np.arange(-0.05,2.1,.1))
    cbar.set_label("Time [yr]")

    min_mix = 5e-1

    for i, yd in enumerate(years_data):

        if (tmin > -1) and (yd["year"] < tmin):
            continue

        if (tmax > -1) and (yd["year"] > tmax):
            continue

        if (tmin == -1) and (i == 0):
            color = 'blue'
        else:
            color = sm.to_rgba(yd["year"])

        p = yd["pressure"]

        # Temperature profile
        ax0.plot(yd["temperature"],p,color=color,lw=lw,alpha=alpha,label="%1.2e"%yd["year"])

        # Mixing ratios
        key = str("mx_"+specie)
        ax1.plot(yd[key],p,label=specie,color=color,lw=lw,alpha=alpha)

        min_mix = min(min_mix,np.amin(yd[key]))

        if plot_init_mx:
            key = str("mv_"+specie)
            if key in yd.keys():
                ax1.scatter(yd[key],p[0], s=40,color='grey')
                ax1.scatter(yd[key],p[0], s=20,color=color)

    ax1.set_xlim([min_mix,1])

    fig.tight_layout()
    fig.savefig(output_dir+"plot_offchem_species_{specie}.pdf")
    plt.close('all')


def plot_offchem_species_entry(handler: Proteus):
    plot_janus_result = True

    # Species to make plots for
    species = ["H2", "H2O", "H", "OH", "O3", "O2",
               "CO2", "CO", "CH4", "C2H2",
               "HCN", "NH3", "NH2", "N2", "NO"]

    # Call plotting function
    for specie in species:
        print(f"Species = {specie}")
        plot_offchem_species(
            output_dir=handler.directories["output"],
            specie=specie,
            tmin=-1,
            plot_init_mx=plot_janus_result,
        )


if __name__ == '__main__':
    print("Plotting offline chemistry (species vs pressure)...")

    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_offchem_species_entry(handler)

    print("Done!")



from __future__ import annotations

from attrs import define, field, validators


@define
class Outgas:
    """Outgassing parameters (fO2) and included volatiles.

    Attributes
    ----------
    fO2_shift_IW: float
        Homogeneous oxygen fugacity in the magma ocean used to represent redox state (log10 units relative to Iron-Wustite).
    module: str
        Outgassing module to be used. Choices: 'calliope', 'atmodeller'.
    calliope: Calliope
        Parameters for CALLIOPE module.
    atmodeller: Atmodeller
        Parameters for atmodeller module.
    """
    fO2_shift_IW: float

    module: str = field(validator=validators.in_(('calliope', 'atmodeller')))

    calliope: Calliope
    atmodeller: Atmodeller


@define
class Calliope:
    """Module parameters for Calliope.

    Attributes
    ----------
    T_floor: float
        Temperature floor applied to outgassing calculation [K].
    include_H2O: bool
        If True, include H2O outgassing.
    include_CO2: bool
        If True, include CO2 outgassing.
    include_N2: bool
        If True, include N2 outgassing.
    include_S2: bool
        If True, include S2 outgassing.
    include_SO2: bool
        If True, include SO2 outgassing.
    include_H2: bool
        If True, include H2 outgassing.
    include_CH4: bool
        If True, include CH4 outgassing.
    include_CO: bool
        If True, include CO outgassing.
    """
    T_floor: float  = field(validator=validators.gt(0.0))
    include_H2O: bool = False
    include_CO2: bool = False
    include_N2: bool = False
    include_S2: bool = False
    include_SO2: bool = False
    include_H2: bool = False
    include_CH4: bool = False
    include_CO: bool = False

    def is_included(self, vol: str) -> bool:
        """Helper method for getting flag if `vol` is included in outgassing."""
        return getattr(self, f'include_{vol}')


@define
class Atmodeller:
    """Module parameters for Atmodeller.

    Attributes
    ----------
    some_parameter: str
        Not used currently.
    """
    some_parameter: str



from __future__ import annotations

from attrs import define, field
from attrs.validators import gt, in_, lt


@define
class Struct:
    """Planetary structure (mass, radius).

    set_by: str
        What defines the interior structure? Choices: "mass_tot", "radius_int".
    mass_tot: float
        Total mass of the planet [M_earth]
    radius_int: float
        Radius of the atmosphere-mantle boundary [R_earth]
    corefrac: float
        Fraction of the planet's interior radius corresponding to the core.
    core_density: float
        Density of the planet's core, optional [kg/m3]
    """

    set_by: str = field(validator=in_(('mass_tot','radius_int')))

    mass_tot: float = field(validator=(gt(0),lt(20)))
    radius_int: float = field(validator=(gt(0),lt(10)))

    corefrac: float = field(validator=(gt(0), lt(1)))
    core_density: float = field(default = 10738.33)



from __future__ import annotations


def none_if_none(val: str) -> str | None:
    """Convert 'none' string into None literal."""
    return None if val == 'none' else val

def dict_replace_none(data):
    """
    Replace all None values with "none" strings.

    Adapted from:
    https://github.com/python-poetry/tomlkit/issues/240#issuecomment-1313283298
    """
    new_data = {}
    for k, v in data.items():
        if isinstance(v, dict):
            v = dict_replace_none(v)
        if v is not None:
            new_data[k] = v
        else:
            new_data[k] = "none"
    return new_data



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)


def plot_observables(hf_all:pd.DataFrame, output_dir: str, plot_format: str="pdf", t0: float=100.0):

    time = np.array(hf_all["Time"] )
    if np.amax(time) < 2:
        log.debug("Insufficient data to make plot_observables")
        return

    log.info("Plot observables")

    # make plot
    lw = 1.2
    scale = 1.1
    fig,axl = plt.subplots(1,1, figsize=(7*scale,4*scale))

    # left axis
    axl.plot(time, hf_all["transit_depth"]*1e6, lw=lw, color='k')
    axl.set_ylabel("Transit depth [ppm]")

    # right axis
    axr = axl.twinx()
    color = "tab:red"
    axr.plot(time, hf_all["contrast_ratio"]*1e9, lw=lw, color=color)
    axr.set_ylabel("Contrast ratio [ppb]")
    axr.yaxis.label.set_color(color)
    axr.tick_params(axis='y', colors=color)

    # x-axis
    axl.set_xlabel("Time [yr]")
    axl.set_xscale("log")
    axl.set_xlim(left=t0, right=np.amax(time))
    axl.grid(alpha=0.2)

    plt.close()
    plt.ioff()

    fpath = os.path.join(output_dir, "plot_observables.%s"%plot_format)
    fig.savefig(fpath, dpi=200, bbox_inches='tight')


def plot_observables_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plot
    plot_observables(
        hf_all=hf_all,
        output_dir=handler.directories["output"],
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_observables_entry(handler)



# Generic atmosphere wrapper
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import pandas as pd
from scipy.integrate import solve_ivp

if TYPE_CHECKING:
    from proteus.config import Config

from proteus.atmos_clim.common import get_spfile_path
from proteus.utils.helper import PrintHalfSeparator, UpdateStatusfile, safe_rm

atm = None
log = logging.getLogger("fwl."+__name__)

def RunAtmosphere(config:Config, dirs:dict, loop_counter:dict,
                  wl: list, fl:list,
                  update_stellar_spectrum:bool, hf_all:pd.DataFrame, hf_row:dict):
    """Run Atmosphere submodule.

    Generic function to run an atmospheric simulation with either JANUS, AGNI or dummy.
    Writes into the hf_row generic variable passed as an arguement.

    Parameters
    ----------
        config : Config
            Configuration options and other variables
        dirs : dict
            Dictionary containing paths to directories
        loop_counter : dict
            Dictionary containing iteration information
        wl : list
            Wavelength [nm]
        fl : list
            Flux at 1 AU [erg s-1 cm-2 nm-1]
        update_stellar_spectrum : bool
            Spectral file path
        hf_all : pd.DataFrame
            Dataframe containing simulation variables (now and historic)
        hf_row : dict
            Dictionary containing simulation variables for current iteration

    """

    #Warning! Find a way to store atm object for AGNI
    global atm

    PrintHalfSeparator()

    # Warnings
    if config.atmos_clim.albedo_pl > 1.0e-9:
        if config.atmos_clim.rayleigh:
            log.warning("Physically inconsistent options selected: "
                        "`albedo_pl > 0` and `rayleigh = True`")
        if config.atmos_clim.cloud_enabled:
            log.warning("Physically inconsistent options selected: "
                        "`albedo_pl > 0` and `cloud_enabled = True`")

    # Handle new surface temperature
    if config.atmos_clim.surf_state == 'mixed_layer':
        hf_row["T_surf"] = ShallowMixedOceanLayer(hf_all.iloc[-1].to_dict(), hf_row)

    elif config.atmos_clim.surf_state == 'fixed':
        hf_row["T_surf"] = hf_row["T_magma"]

    # elif surf_state=='skin':
    #    Don't do anything here, because this will be handled by the atmosphere model.

    if config.atmos_clim.module == 'janus':
        # Import
        from proteus.atmos_clim.janus import InitAtm, InitStellarSpectrum, RunJANUS

        # Run JANUS
        no_atm = bool(atm is None)
        #Enforce surface temperature at first iteration
        if no_atm:
            hf_row["T_surf"] = hf_row["T_magma"]

        #Init atm object if first iteration or change in stellar spectrum
        if no_atm or update_stellar_spectrum:
            spectral_file_nostar = get_spfile_path(dirs["fwl"], config)
            if not os.path.exists(spectral_file_nostar):
                UpdateStatusfile(dirs, 20)
                raise Exception("Spectral file does not exist at '%s'" % spectral_file_nostar)
            InitStellarSpectrum(dirs, wl, fl, spectral_file_nostar)
            atm = InitAtm(dirs, config)

        atm_output = RunJANUS(atm, dirs, config, hf_row, hf_all)

    elif config.atmos_clim.module == 'agni':
        # Import
        from proteus.atmos_clim.agni import (
            activate_julia,
            deallocate_atmos,
            init_agni_atmos,
            run_agni,
            update_agni_atmos,
        )

        # Run AGNI
        # Initialise atmosphere struct
        spfile_path = os.path.join(dirs["output"] , "runtime.sf")
        no_atm = bool(atm is None)
        if no_atm or update_stellar_spectrum:
            log.debug("Initialise new atmosphere struct")

            # first run?
            if no_atm:
                activate_julia(dirs)
                # surface temperature guess
                hf_row["T_surf"] = hf_row["T_magma"]
            else:
                # Remove old spectral file if it exists
                safe_rm(spfile_path)
                safe_rm(spfile_path+"_k")

                # deallocate old atmosphere
                deallocate_atmos(atm)

            # allocate new
            atm = init_agni_atmos(dirs, config, hf_row)

            # Check allocation was ok
            if not bool(atm.is_alloc):
                log.error("Failed to allocate atmos struct")
                UpdateStatusfile(dirs, 22)
                exit(1)

        # Update profile
        atm = update_agni_atmos(atm, hf_row, dirs)

        # Run solver
        atm, atm_output = run_agni(atm, loop_counter["total"], dirs, config, hf_row)

    elif config.atmos_clim.module == 'dummy':
        # Import
        from proteus.atmos_clim.dummy import RunDummyAtm
        # Run dummy atmosphere model
        atm_output = RunDummyAtm(dirs, config, hf_row["T_magma"], hf_row["F_ins"],
                                    hf_row["R_int"], hf_row["M_int"], hf_row["P_surf"])

    # Store atmosphere module output variables
    hf_row["rho_obs"]= atm_output["rho_obs"] # [kg m-3]
    hf_row["p_obs"]  = atm_output["p_obs"]   # [bar]
    hf_row["z_obs"]  = atm_output["z_obs"]   # [m]
    hf_row["F_atm"]  = atm_output["F_atm"]
    hf_row["F_olr"]  = atm_output["F_olr"]
    hf_row["F_sct"]  = atm_output["F_sct"]
    hf_row["T_surf"] = atm_output["T_surf"]
    hf_row["F_net"]  = hf_row["F_int"] - hf_row["F_atm"]
    hf_row["bond_albedo"]= atm_output["albedo"]
    hf_row["p_xuv"]  = atm_output["p_xuv"]                  # Closest pressure from Pxuv    [bar]
    hf_row["z_xuv"]  = atm_output["z_xuv"]                  # Height at Pxuv                [m]
    hf_row["R_xuv"]  = hf_row["R_int"] + hf_row["z_xuv"]    # Radius at z_xuv [m]

    # Calculate observables (measured at infinite distance)
    R_obs = hf_row["z_obs"] + hf_row["R_int"] # observed radius [m]
    hf_row["transit_depth"] =  (R_obs / hf_row["R_star"])**2.0
    hf_row["contrast_ratio"] = ((hf_row["F_olr"]+hf_row["F_sct"])/hf_row["F_ins"]) * \
                                 (R_obs / hf_row["separation"])**2.0

def ShallowMixedOceanLayer(hf_cur:dict, hf_pre:dict):

    # This scheme is not typically used, but it maintained here from legacy code
    # We could consider removing it in the future.

    log.info(">>>>>>>>>> Flux convergence scheme <<<<<<<<<<<")

    # For SI conversion
    yr          = 3.154e+7      # s

    # Last T_surf and time from atmosphere, K
    t_cur  = hf_cur["Time"]*yr
    t_pre  = hf_pre["Time"]*yr
    Ts_pre = hf_pre["T_surf"]

    # Properties of the shallow mixed ocean layer
    c_p_layer   = 1000          # J kg-1 K-1
    rho_layer   = 3000          # kg m-3
    depth_layer = 1000          # m

    def ocean_evolution(t, y):
        # Specific heat of mixed ocean layer
        mu      = c_p_layer * rho_layer * depth_layer # J K-1 m-2
        # RHS of ODE
        RHS     = - hf_cur["F_net"] / mu
        return RHS

    # Solve ODE
    sol_curr  = solve_ivp(ocean_evolution, [t_pre, t_cur], [Ts_pre])

    # New current surface temperature from shallow mixed layer
    Ts_cur = sol_curr.y[0][-1] # K

    return Ts_cur



from __future__ import annotations

import sys

from proteus import Proteus


def get_handler_from_argv(default: str = 'init_coupler.toml') -> Proteus:
    """Attempts to get handler from argv, otherwise defaults to default"""
    config_path = sys.argv[1] if len(sys.argv) == 2 else default
    handler = Proteus(config_path=config_path)

    return handler



from __future__ import annotations

from attrs import define, field
from attrs.validators import ge, gt, in_, le

from ._converters import none_if_none


@define
class Star:
    """Stellar parameters, model selection.

    You can find useful reference data in the [documentation](https://fwl-proteus.readthedocs.io/en/latest/data/#stars).

    Attributes
    ----------
    mass: float
        Stellar mass [M_sun]. Note that for Mors,
        it should be between 0.1 and 1.25 solar masses.
        Values outside of the valid range will be clipped.
    age_ini: float
        Age of system at model initialisation [Gyr].
    module: str | None
        Select star module to use.
    mors: Mors
        Parameters for MORS module.
    dummy: StarDummy
        Parameters for the dummy star module
    """
    mass: float = field(validator=(ge(0.1), le(1.25)))
    age_ini: float = field(validator=gt(0))

    module: str | None = field(
        validator=in_((None, 'mors', 'dummy')),
        converter=none_if_none,
    )

    mors: Mors
    dummy: StarDummy


@define
class Mors:
    """Module parameters for MORS module.

    Attributes
    ----------
    rot_pctle: float
        Rotation rate, as a percentile of stellar population with the same mass [%].
    tracks: str
        Stellar evolution track to be used. Choices: 'spada', 'baraffe'.
    age_now: float
        Observed estimated age of the star [Gyr].
    spec: str
        Name of file containing stellar spectrum. See [documentation](https://fwl-proteus.readthedocs.io/en/latest/data/#stars) for potential file names.
    """
    rot_pctle: float = field(validator=(ge(0), le(100)))
    tracks: str = field(validator=in_(('spada', 'baraffe')))
    age_now: float = field(validator=gt(0))
    spec: str

@define
class StarDummy:
    """Dummy star module.

    Attributes
    ----------
    radius: float
        Observed radius [R_sun].
    Teff: float
        Observed effective temperature [K].
    """
    radius: float = field(validator=gt(0))
    Teff: float = field(validator=gt(0))



from __future__ import annotations

import glob
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from proteus.utils.plot import get_colour, latexify
from proteus.utils.plot_offchem import offchem_read_year

if TYPE_CHECKING:
    from proteus import Proteus

mpl.use("Agg")


def plot_offchem_year(
        output_dir: str,
        year_dict: dict,
        species: list,
        plot_init_mx: bool=False,
    ):
    """Plot a set of species in the atmosphere, for a single year only

    Reads-in the data from output_dir for a single year. Can also
    include JANUS/SPIDER mixing ratios as dashed lines, which were used to
    initialise each VULCAN run.

    Parameters
    ----------
    output_dir : str
        Output directory that was specified in the PROTEUS cfg file
    species : list
        Which species to plot? (e.g. H2O) If none, then plot all.
    tmin : float
        Initial year to include [yr]
    tmax : float
        Final yeat to include [yr]
    plot_init_mx : bool
        Include initial mixing ratios for each VULCAN run in plot?
    """

    if species is None:
        species = []
        for k in year_dict.keys():
            if "_" not in k:
                continue
            species.append(k.split("_")[1])

    lw = 2

    mpl.use("Agg")

    fig, (ax0,ax1) = plt.subplots(1,2,sharey=True,figsize=(8,4))

    # Temperature profile
    ax0.set_title("Time = %g kyr" % (year_dict["year"]/1.e3))
    ax0.set_ylabel("Pressure [bar]")
    ax0.invert_yaxis()
    ax0.set_yscale("log")
    ax0.set_xlabel("Temperature [K]")

    ax0.plot(year_dict["temperature"],year_dict["pressure"],color='black',lw=lw)

    # Mixing ratios
    ax1.set_xlabel("Mole fraction")
    ax1.set_xscale("log")


    min_mix = 5e-1
    for i,s in enumerate(species):

        ls = 'solid'
        color = get_colour(s)

        # VULCAN result
        key = str("mx_"+s)
        if key in year_dict.keys():
            line = ax1.plot(year_dict[key],year_dict["pressure"],label=latexify(s),ls=ls,lw=lw,color=color)[0]
            color = line.get_color()
            min_mix = min(min_mix,np.amin(year_dict[key]))

        # JANUS result
        if plot_init_mx:
            key = str("mv_"+s)
            if key in year_dict.keys():
                ax1.scatter([year_dict[key]],[np.amax(year_dict["pressure"])],color=color)

    ax1.set_xlim([min_mix,1])
    ax1.legend(loc="lower left")

    fig.tight_layout()
    fig.savefig(output_dir+"plot_offchem_year_%d.pdf"%year_dict["year"])
    plt.close(fig)


def plot_offchem_year_entry(handler: Proteus, **kwargs):
    plot_janus = True

    # Species to make plots for
    species = ["H2", "H2O", "H", "OH", "CO2", "CO", "CH4","HCN", "NH3", "N2", "NO"]

    # Load data
    ls = glob.glob(handler.directories["output"]+"offchem/*/output.vul")
    years = [int(f.split("/")[-2]) for f in ls]
    years_data = [offchem_read_year(handler.directories["output"],y,read_const=plot_janus) for y in years]

    if len(years) == 0:
        raise Exception('No VULCAN output files found')

    # Call plotting function
    for yd in years_data:
        print("Year = %d" % yd["year"])
        plot_offchem_year(
            output_dir=handler.directories["output"],
            year_dict=yd,
            species=species,
            plot_init_mx=plot_janus,
        )


if __name__ == '__main__':
    print("Plotting offline chemistry for each year (mixing ratios vs pressure)...")

    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_offchem_year_entry(handler)

    print("Done!")



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd

from proteus.utils.constants import gas_list
from proteus.utils.plot import get_colour, latexify

if TYPE_CHECKING:
    from proteus import Proteus
    from proteus.config import Config

log = logging.getLogger("fwl."+__name__)

def plot_global(hf_all: pd.DataFrame, output_dir: str, config: Config,
                logt: bool=True, tmin: float=1e1):

    if np.amax(hf_all["Time"]) < 3:
        log.debug("Insufficient data to make plot_global")
        return
    hf = hf_all.loc[hf_all["Time"]>0]

    log.info("Plot global")

    # Plotting parameters
    lw=2.0
    al=0.95
    fig_ratio=(3,2)
    fig_scale=4.0
    leg_kwargs = {
        "frameon":1,
        "fancybox":True,
        "framealpha":0.9
    }


    #    Volatile parameters (keys=vols, vals=quantites_over_time)
    vol_present = {} # Is present ever? (true/false)
    vol_vmr     = {} # Volume mixing ratio
    vol_bars    = {} # Surface partial pressure [bar]
    vol_mol_atm = {} # Moles in atmosphere
    vol_mol_int = {} # Moles in interior
    vol_mol_tot = {} # Moles in total
    vol_intpart = {} # Partitioning into int

    for vol in gas_list:
        # Check vmr for presence
        this_vmr = np.array(hf[vol+"_vmr"])
        vol_present[vol] = True

        if (np.amax(this_vmr) < 1.0e-20) or not config.outgas.calliope.is_included(vol):
            vol_present[vol] = False
            continue
        vol_vmr[vol] = this_vmr

        # Do other variables
        vol_bars[vol]    = np.array(hf[vol+"_bar"])
        vol_mol_atm[vol] = np.array(hf[vol+"_mol_atm"])
        vol_mol_tot[vol] = np.array(hf[vol+"_mol_total"])
        vol_mol_int[vol] = vol_mol_tot[vol] - vol_mol_atm[vol]

        # Volatile partitioning into the interior
        # Requires special treatment for when moles=0, which occurs when atmosphere escapes.
        mask = np.argwhere(vol_mol_tot[vol] < 1e-10) # mask of values where moles=0
        vol_mol_int[vol][mask] = 0
        vol_mol_tot[vol][mask] = 1
        vol_intpart[vol] = vol_mol_int[vol]/vol_mol_tot[vol]

    # Init plot
    fig,axs = plt.subplots(3,2, figsize=(fig_ratio[0]*fig_scale, fig_ratio[1]*fig_scale), sharex=True)
    ax_tl = axs[0][0]
    ax_cl = axs[1][0]
    ax_bl = axs[2][0]
    ax_tr = axs[0][1]
    ax_cr = axs[1][1]
    ax_br = axs[2][1]
    axs = (ax_tl, ax_cl, ax_bl, ax_tr, ax_cr, ax_br) # flatten

    # Set y-labels
    ax_tl.set_ylabel('Upward flux [W m$^{-2}$]')
    ax_cl.set_ylabel(r'$T_\mathrm{s}$ [K]')
    ax_bl.set_ylabel('Planet fraction')
    ax_tr.set_ylabel(r'$p_{\mathrm{i}}$ [bar]')
    ax_cr.set_ylabel(r'$\chi^{\mathrm{atm}}_{\mathrm{i}}$ [%]')
    ax_br.set_ylabel(r'$m^{\mathrm{int}}_{\mathrm{i}}/m^{\mathrm{tot}}_{\mathrm{i}}$ [%]')

    # Set x-labels
    for ax in (ax_bl, ax_br):
        ax.set_xlabel("Time [yr]")

    # Set titles
    ax_titles = [
        "(a) Net heat flux to space",
        "(c) Surface temperature",
        "(e) Mantle evolution",
        "(b) Surface gas partial pressure",
        "(d) Surface gas mole fraction",
        "(f) Interior volatile partitioning"
    ]
    for i,ax in enumerate(axs):
        ax.text(0.011, 0.015, ax_titles[i],
                transform=ax.transAxes,
                horizontalalignment='left',
                verticalalignment='bottom',
                fontsize=11,
                zorder=20,
                bbox=dict(fc='white', ec="white", alpha=0.5, pad=0.1, boxstyle='round')
                )

    # Move right subplots y-axes to right side
    for ax in (ax_tr, ax_cr, ax_br):
        ax.yaxis.set_label_position("right")
        ax.yaxis.tick_right()

    # Percentage plots
    for ax in (ax_cr, ax_br):
        ax.yaxis.set_major_locator(ticker.MultipleLocator(20))
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(10))

    # Log axes
    if logt:
        for ax in axs:
            ax.set_xscale("log")
    ax_tl.set_yscale("symlog", linthresh=0.1)
    ax_tr.set_yscale("log")

    # Set xlim
    xmin = max(tmin, 1.0)
    if logt:
        xmax = max(1.0e6,np.amax(hf["Time"]))
        xlim = (xmin,10 ** np.ceil(np.log10(xmax*1.1)))
    else:
        xmax = np.amax(hf["Time"])
        xlim = (1.0, xmax)
    for ax in axs:
        ax.set_xlim(xlim[0],  max(xlim[1], xlim[0]+1))


    # PLOT ax_tl
    ax_tl.plot( hf["Time"], hf["F_int"],  color=get_colour("int"),  lw=lw, alpha=al,  label="Int.", linestyle='dashed')
    ax_tl.plot( hf["Time"], hf["F_atm"],  color=get_colour("atm"),  lw=lw, alpha=al,  label="Atm."  )
    ax_tl.plot( hf["Time"], hf["F_olr"],  color=get_colour("OLR"),  lw=lw, alpha=al,  label="OLR"   )
    ax_tl.plot( hf["Time"], hf["F_tidal"], color=get_colour("tidal"), lw=lw, alpha=al,  label="Tidal")
    ax_tl.plot( hf["Time"], hf["F_radio"], color=get_colour("radio"), lw=lw, alpha=al,  label="Radio.")
    ax_tl.legend(loc='center left', **leg_kwargs)
    ymin, ymax = 0.0, 100.0
    for k in ("F_int","F_atm","F_olr","F_tidal","F_radio"):
        ymin = min(ymin, np.amin(hf[k]))
        ymax = max(ymax, np.amax(hf[k]))
    ax_tl.set_ylim(bottom=ymin/1.5, top=ymax*1.5)

    # PLOT ax_cl
    min_temp = np.amin(hf["T_surf"])
    max_temp = np.amax(hf["T_surf"])
    ax_cl.plot(hf["Time"], hf["T_surf"], ls="dashed", lw=lw, alpha=al, color=get_colour("int"))
    ax_cl.plot(hf["Time"], hf["T_surf"], ls="-",      lw=lw, alpha=al, color=get_colour("atm"))
    ax_cl.set_ylim(min(1000.0,min_temp-25) , max(3500.0,max_temp+25))

    # PLOT ax_bl
    ax_bl.axhline( y=config.struct.corefrac,     ls='dashed', lw=lw*1.5, alpha=al, color=get_colour("core"), label=r'C-M boundary' )
    ax_bl.plot( hf["Time"], 1.0-hf["RF_depth"],   color=get_colour("int"), ls="solid",    lw=lw, alpha=al, label=r'Rheol. front')
    ax_bl.plot( hf["Time"],     hf["Phi_global"], color=get_colour("atm"), linestyle=':', lw=lw, alpha=al, label=r'Melt fraction')

    ax_bl.legend(loc='center left', **leg_kwargs)
    ax_bl.set_ylim(0.0,1.01)

    # PLOT ax_tr
    ax_tr.plot( hf["Time"], hf["P_surf"], color='black', linestyle='dashed', lw=lw*1.5, label=r'Total')
    bar_min, bar_max = 0.1, 10.0
    bar_max = max(bar_max, np.amax(hf["P_surf"]))
    for vol in gas_list:
        if not vol_present[vol]:
            continue
        ax_tr.plot( hf["Time"], vol_bars[vol], color=get_colour(vol), lw=lw, alpha=al, label=latexify(vol))
        bar_min = min(bar_min, np.amin(vol_bars[vol]))
    ax_tr.set_ylim(max(1.0e-7,min(bar_min, 1.0e-1)), bar_max * 2.0)
    ax_tr.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=5) )

    # PLOT ax_cr
    for vol in gas_list:
        if not vol_present[vol]:
            continue
        ax_cr.plot( hf["Time"], vol_vmr[vol]*100.0, color=get_colour(vol), lw=lw, alpha=al, label=latexify(vol))
    ax_cr.set_ylim(0, 101)

    # PLOT ax_br
    for vol in gas_list:
        if not vol_present[vol]:
            continue
        ax_br.plot( hf["Time"], vol_intpart[vol]*100.0, color=get_colour(vol), lw=lw, alpha=al, label=latexify(vol))
    ax_br.set_ylim(0,101)
    ax_br.legend(loc='center left', ncol=2, **leg_kwargs).set_zorder(20)

    # Save plot
    fig.subplots_adjust(wspace=0.05,hspace=0.1)
    plt_name = "plot_global"
    if logt:
        plt_name += "_log"
    else:
        plt_name += "_lin"

    fig.savefig(output_dir+"/%s.%s"%(plt_name,config.params.out.plot_fmt),
                bbox_inches='tight', dpi=200)


def plot_global_entry(handler: Proteus):
    # read helpfile
    hf_all = pd.read_csv(os.path.join(handler.directories['output'], "runtime_helpfile.csv"), sep=r"\s+")

    # make plot
    for logt in [True,False]:
        plot_global(
            hf_all=hf_all,
            output_dir=handler.directories['output'],
            config=handler.config,
            logt=logt,
            tmin=1e1,
        )

if __name__ == "__main__":
    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_global_entry(handler)



from __future__ import annotations

import glob
import logging
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from proteus.utils.constants import const_c, const_h, const_k
from proteus.utils.helper import natural_sort

if TYPE_CHECKING:
    from proteus import Proteus

log = logging.getLogger("fwl."+__name__)

def planck_function(lam, T):
    """Plots stellar flux from `output/` versus time (colorbar)

    Planck function value at stellar surface
    lam in nm
    erg s-1 cm-2 nm-1
    """

    x = lam * 1.0e-9   # convert nm -> m
    hc_by_kT = const_h*const_c / (const_k*T)

    planck_func = 1.0/( (x ** 5.0) * ( np.exp( hc_by_kT/ x) - 1.0 ) )
    planck_func *= 2 * const_h * const_c * const_c #  w m-2 sr-1 s-1 m-1
    planck_func *= np.pi * 1.0e3 * 1.0e-9  # erg s-1 cm-2 nm-1

    return planck_func


def plot_sflux(output_dir: str, wl_max: float = 6000.0,
                plt_modern:bool=True, plot_format: str="pdf"):
    """Plots stellar flux vs time for all wavelengths

    Note that this function will plot the flux from EVERY file it finds.
    Saves plot as 'cpl_sflux.pdf' in  the output directory.

    Parameters
    ----------
    output_dir : str
        Directory for both reading from and saving to.

    wl_max : float
        Upper limit of wavelength axis [nm]
    plt_modern : bool
        Include modern spectrum in plot?
    plot_format: str
        Output figure file format

    """

    mpl.use('Agg')
    star_cmap = plt.get_cmap('Spectral')

    # Find and sort files
    files_unsorted = glob.glob(output_dir+"/data/*.sflux")
    files = natural_sort(files_unsorted)

    if (len(files) == 0):
        log.warning("No files found when trying to plot stellar flux")
        return

    log.info("Plot stellar flux")

    # Downsample data
    if (len(files) > 200):
        files = files[::2]
    if (len(files) > 500):
        files = files[::2]
    if (len(files) > 1000):
        files = files[::3]

    if (len(files) != len(files_unsorted)):
        log.debug("Downsampled data over time")

    # Arrays for storing data over time
    time_t = []
    wave_t = []
    flux_t = []
    for f in files:
        # Load data
        X = np.loadtxt(f,skiprows=1,delimiter='\t').T

        # Parse data
        time = int(f.split('/')[-1].split('.')[0])
        if time < 0:
            continue

        wave = X[0]
        flux = X[1]

        # Save data
        time_t.append(time)
        wave_t.append(wave)
        flux_t.append(flux)

    time_t = np.array(time_t)
    wave_t = np.array(wave_t)
    flux_t = np.array(flux_t)

    # Create figure
    N = len(time_t)

    fig,ax = plt.subplots(1,1)

    # Colorbar
    justone = bool(N == 1)
    if not justone:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes('right', size='3%', pad=0.05)

        vmin = max(time_t[0],1.0)
        vmax = time_t[-1]
        norm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
        sm = plt.cm.ScalarMappable(cmap=star_cmap, norm=norm)
        sm.set_array([])
        cbar = fig.colorbar(sm, cax=cax, orientation='vertical')
        cbar.set_label("Time [yr]")
    else:
        log.warning("Only one spectrum was found")

    ax.set_yscale("log")
    ax.set_ylabel(r"Flux [erg / (s cm$^2$ nm)]")

    ax.set_xscale("log")
    ax.set_xlabel("Wavelength [nm]")
    ax.set_xlim([0.5,max(1.0,wl_max)])
    ax.set_title("TOA flux versus wavelength")

    # Plot historical spectra
    for i in range(N):
        if justone:
            c = 'tab:blue'
            label = "%.2e yr"%(time_t[i])
        else:
            c = sm.to_rgba(time_t[i])
            label = None
        ax.plot(wave_t[i],flux_t[i],color=c,lw=0.7,alpha=0.6, label=label)

    # Plot current spectrum (use the copy made in the output directory)
    if plt_modern:
        X = np.loadtxt(output_dir+'/-1.sflux',skiprows=2).T
        ax.plot(X[0],X[1],color='black',label='Modern',lw=0.8,alpha=0.9)

    if plt_modern or justone:
        ax.legend()

    plt.close()
    plt.ioff()
    fig.savefig(output_dir+"/plot_sflux.%s"%plot_format,
                bbox_inches='tight', dpi=200)


def plot_sflux_entry(handler: Proteus):
    plot_sflux(
        output_dir=handler.directories['output'],
        plt_modern=handler.config.star.module == "mors",
        plot_format=handler.config.params.out.plot_fmt,
    )


if __name__ == '__main__':
    print("Plotting stellar flux over time (colorbar)...")

    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_sflux_entry(handler)

    print("Done!")



# Setup logger for PROTEUS
from __future__ import annotations

import logging
import os
import sys


# Fake file-like stream object that redirects writes to a logger instance.
class StreamToLogger(object):
    # https://stackoverflow.com/a/36296215
    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ''
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == '\n':
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != '':
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ''

class CustomFormatter(logging.Formatter):

    part1 = "[\033["

    info =  "32"
    warn =  "93"
    debug = "96"
    error = "91"

    part2 = "m\033[1m %(levelname)-5s \033[21m\033[0m] %(message)s"

    FORMATS = {
        logging.DEBUG:   part1+debug+part2,
        logging.INFO:    part1+info +part2,
        logging.WARNING: part1+warn +part2,
        logging.ERROR:   part1+error+part2,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

# Custom logger instance
def setup_logger(logpath:str="new.log",level:str="INFO",logterm:bool=True):

    logger_name = "fwl"

    # https://stackoverflow.com/a/61457119
    custom_logger = logging.getLogger(logger_name)

    if os.path.exists(logpath):
        os.remove(logpath)

    level = str(level).strip().upper()
    if level not in ["INFO", "DEBUG", "ERROR", "WARNING"]:
        raise ValueError(f"Invalid log level: {level}")
    level_code = logging.getLevelName(level)

    # Add terminal output to logger
    if logterm:
        sh = logging.StreamHandler(sys.stdout)
        sh.setFormatter(CustomFormatter())
        sh.setLevel(level_code)
        custom_logger.addHandler(sh)

    # Add file output to logger
    fh = logging.FileHandler(logpath)
    fh.setFormatter(logging.Formatter("[ %(levelname)-5s ] %(message)s"))

    fh.setLevel(level)
    custom_logger.addHandler(fh)
    custom_logger.setLevel(level_code)

    # Capture unhandled exceptions
    # https://stackoverflow.com/a/16993115
    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            custom_logger.error("KeyboardInterrupt")
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        custom_logger.critical("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
    sys.excepthook = handle_exception

    return custom_logger

def GetCurrentLogfileIndex(output_dir:str):
    '''
    Get the index of the current logfile, returning -1 if none exists
    '''
    i=0
    j=-1
    while i<99:
        fname = "proteus_%02d.log"%i
        fpath = os.path.join(output_dir, fname)

        if os.path.exists(fpath):
            j=i
            i += 1
        else:
            break

    return j

def GetLogfilePath(output_dir:str, j:int):
    '''
    Get path to logfile with index j
    '''

    if j>99:
        raise Exception("Cannot create logfile - too many in output folder already")

    return os.path.join(output_dir, "proteus_%02d.log"%j)



from __future__ import annotations

import glob
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from proteus.utils.helper import find_nearest
from proteus.utils.plot import get_colour, latexify
from proteus.utils.plot_offchem import offchem_read_year

if TYPE_CHECKING:
    from proteus import Proteus


def plot_offchem_time(
        output_dir: str,
        species: list,
        plot_init_mx: bool=False,
        tmin: int=-1,
        prange: list | float | None=None
    ):
    """Plot evolution of chemistry according to offline VULCAN output.

    Reads-in the data from output_dir for a set of provided species. Can also
    include JANUS/SPIDER mixing ratios as dashed lines, which were used to
    initialise each VULCAN run. Mixing ratios are averaged over the provided
    pressure range.

    Parameters
    ----------
    output_dir : str
        Output directory that was specified in the PROTEUS cfg file
    species : list
        List of species to plot
    prange : list(2) or float or None
        Minimum and maximum pressures (bar) to average over. If None, then whole column is used. If float, then single layer is used.
    plot_init_mx : bool
        Include initial mixing ratios for each VULCAN run in plot?
    tmin : float
        Initial plotting time (-1 for start of data)
    """

    mpl.use("Agg")

    ls = glob.glob(output_dir+"/offchem/*/output.vul")
    years = [int(f.split("/")[-2]) for f in ls]
    years_data = [offchem_read_year(output_dir,y,read_const=plot_init_mx) for y in years]

    if len(years) == 0:
        raise Exception('No VULCAN output files found')

    lw = 2

    fig, ax = plt.subplots(1,1,figsize=(8,4))

    ax.set_ylabel("Mean mixing ratio")
    ax.set_yscale("log")
    ax.set_xlabel("Time")
    ax.set_xlabel("Time [yr]")
    ax.set_xscale("log")
    title_str = "Mixing ratio versus time "

    prange_mode = -1
    if prange is None:
        prange_mode = 0
        title_str += "(whole column)"
    elif isinstance(prange, float):
        prange_mode = 1
        title_str += str("(p ≈ %1.1e bar)" % prange)
    elif len(prange) == 2:
        prange_mode = 2
        title_str += str("(p ≈ [%1.1e,%1.1e] bar)" % (prange[0],prange[1]))
    else:
        raise Exception("Error parsing value for parameter 'prange': " + str(prange))


    for i,sp in enumerate(species):

        color = get_colour(sp)

        times =     []
        mx_vul =    []
        mx_aeo =    []


        for yd in years_data:
            times.append(yd["year"])

            clean_mx = np.nan_to_num(yd["mx_"+sp])

            # Crop to pressure range
            if prange_mode == 0:
                mean_mx = np.mean(clean_mx)
            elif prange_mode == 1:
                _,vi = find_nearest(yd["pressure"],prange)
                mean_mx = clean_mx[vi]
            elif prange_mode == 2:
                vmin,vimin = find_nearest(yd["pressure"],prange[0])
                vmax,vimax = find_nearest(yd["pressure"],prange[1])
                if vimin == vimax:
                    raise Exception("Upper and lower bound of 'prange' are too close!")
                if vimin > vimax:
                    vimin,vimax = vimax,vimin
                    vmin,vmax = vmax,vmin
                mean_mx = np.mean(clean_mx[vimin:vimax])

            mean_mx = np.mean(clean_mx)
            mx_vul.append(float(mean_mx))

        ax.plot(times,mx_vul,color=color,label=latexify(sp),lw=lw,zorder=4)

        if plot_init_mx:
            if str("mv_"+sp) in years_data[0].keys():
                for yd in years_data:
                    clean_mx = np.nan_to_num(yd["mv_"+sp])
                    mean_mx = np.mean(clean_mx)
                    mx_aeo.append(float(mean_mx))

                ax.plot(times,mx_aeo,color=color,linestyle='--',lw=lw,zorder=2)

    title_str += str(", %d samples" % len(years_data))

    ax.set_title(title_str)
    ax.legend(loc="lower left",ncol=6)
    ax.set_ylim([1e-10,1.5])

    if tmin == -1:
        vmin = times[0]
    else:
        vmin = tmin
    ax.set_xlim([max(vmin,1.0),times[-1]*1.2])

    ax.axvline(x=times[-1],color='black',lw=0.7,zorder=1)

    fig.tight_layout()
    fig.savefig(output_dir+"/plot_offchem_time.pdf")
    plt.close('all')


def plot_offchem_time_entry(handler: Proteus):
    # Species to plot
    species = ["H2", "H2O", "H", "OH", "CO2", "CO", "CH4", "HCN", "NH3", "N2", "NO"]

    # Call plotting function
    prange = [1e0, 1e-4]
    plot_offchem_time(
        output_dir=handler.directories["output"],
        species=species,
        tmin=1e3,
        prange=prange,
    )


if __name__ == '__main__':
    print("Plotting offline chemistry (mixing ratios vs time)...")

    from proteus.plot._cpl_helpers import get_handler_from_argv
    handler = get_handler_from_argv()
    plot_offchem_time_entry(handler)

    print("Done!")
